#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
import uuid
from .base_facility import BaseFacilityState
import plotly.graph_objects as go
import plotly.figure_factory as ff
import datetime


class BaseFactory(object, metaclass=abc.ABCMeta):
    """
    BaseFactory class for expressing factory including facilities in a project.
    This class will be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this factory.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None -> str(uuid.uuid4()).
        facility_list (List[BaseFacility], optional):
            Basic parameter.
            List of BaseFacility who belong to this factory.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted BaseTasks.
            Defaults to None -> [].
        parent_factory (BaseFactory, optional):
            Basic parameter.
            Parent factory of this factory.
            Defaults to None.
        max_space_size (float, optional):
            Basic parameter
            Max size of space for placing components
            Default to None -> 1.0
        placed_component_list (List[BaseComponent], optional):
            Basic variable.
            Components which places to this factory in simulation.
            Defaults to None -> [].
        placed_component_id_record(List[List[str]], optional):
            Basic variable.
            Record of placed components ID in simulation.
            Defaults to None -> [].
        cost_list (List[float], optional):
            Basic variable.
            History or record of this factory's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        facility_list=None,
        targeted_task_list=None,
        parent_factory=None,
        max_space_size=None,
        # Basic variables
        cost_list=None,
        placed_component_list=None,
        placed_component_id_record=None,
    ):

        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())

        self.facility_list = facility_list if facility_list is not None else []
        for facility in self.facility_list:
            if facility.factory_id is None:
                facility.factory_id = self.ID

        self.targeted_task_list = (
            targeted_task_list if targeted_task_list is not None else []
        )
        self.parent_factory = parent_factory if parent_factory is not None else None
        self.max_space_size = max_space_size if max_space_size is not None else 1.0

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        if cost_list is not None:
            self.cost_list = cost_list
        else:
            self.cost_list = []

        if placed_component_list is not None:
            self.placed_component_list = placed_component_list
        else:
            self.placed_component_list = []

        if placed_component_id_record is not None:
            self.placed_component_id_record = placed_component_id_record
        else:
            self.placed_component_id_record = []

    def set_parent_factory(self, parent_factory):
        """
        Set parent factory

        Args:
            parent_factory (BaseFactory):
                Parent factory
        Examples:
            >>> t = BaseFactory('t')
            >>> t1 = BaseFactory('t1')
            >>> t.set_parent_factory(t1)
            >>> print(t.parent_factory.name)
            't1'
        """
        self.parent_factory = parent_factory

    def add_facility(self, facility):
        """
        Add facility to self.facility_list

        Args:
            facility (BaseFacility):
                Facility which is added to this factory
        """
        facility.factory_id = self.ID
        self.facility_list.append(facility)

    def get_total_workamount_skill(self, task_name, error_tol=1e-10):
        """
        Get total number of workamount skill of all facilities
        by checking workamount_skill_mean_map.

        Args:
            task_name (str):
                Task name
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.

        Returns:
            float: total workamount skill of target task name
        """
        sum_skill_point = 0.0
        for facility in self.facility_list:
            if facility.has_workamount_skill(task_name, error_tol=error_tol):
                sum_skill_point += facility.workamount_skill_mean_map[task_name]
        return sum_skill_point

    def extend_targeted_task_list(self, targeted_task_list):
        """
        Extend the list of targeted tasks

        Args:
            targeted_task_list (list[BaseTask]):
                List of targeted tasks
        Examples:
            >>> factory = BaseFactory('factory')
            >>> print([targeted_t.name for targeted_t in factory.targeted_task_list])
            []
            >>> factory.extend_targeted_task_list([BaseTask('t1'),BaseTask('t2')])
            >>> print([targeted_t.name for targeted_t in factory.targeted_task_list])
            ['t1', 't2']
        """
        for targeted_task in targeted_task_list:
            self.append_targeted_task(targeted_task)

    def append_targeted_task(self, targeted_task):
        """
        Append targeted task

        Args:
            targeted_task (BaseTask):
                Targeted task
        Examples:
            >>> factory = BaseFactory('factory')
            >>> print([targeted_t.name for targeted_t in factory.targeted_task_list])
            []
            >>> t1 = BaseTask('t1')
            >>> factory.append_targeted_task(t1)
            >>> print([targeted_t.name for targeted_t in factory.targeted_task_list])
            ['t1']
            >>> print([target_f.name for target_f in t1.allocated_factory_list])
            ['factory']
        """
        self.targeted_task_list.append(targeted_task)
        targeted_task.allocated_factory_list.append(self)

    def set_placed_component(self, placed_component):
        """
        Set the placed_factory

        Args:
            placed_component (BaseComponent):
        """
        self.placed_component_list.append(placed_component)

    def remove_placed_component(self, placed_component):
        self.placed_component_list.remove(placed_component)

    def can_put(self, component, error_tol=1e-8):
        """
        Check whether the target component can be put to this factory in this time

        Args:
            component (BaseComponent):
                BaseComponent
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-8.

        Returns:
            bool: whether the target component can be put to this factory in this time
        """
        can_put = False
        sum_space_size = sum([c.space_size for c in self.placed_component_list])
        if sum_space_size + component.space_size <= self.max_space_size + error_tol:
            can_put = True
        return can_put

    def initialize(self):
        """
        Initialize the changeable variables of BaseFactory

        - cost_list
        - placed_component_list
        - placed_component_id_record
        - changeable basic variable of BaseFacility in facility_list
        """
        self.cost_list = []
        self.placed_component_list = []
        self.placed_component_id_record = []
        for w in self.facility_list:
            w.initialize()

    def add_labor_cost(self, only_working=True, add_zero_to_all_facilities=False):
        """
        Add labor cost to resources in this factory.

        Args:
            only_working (bool, optional):
                If True, add labor cost to only WORKING facilities in this factory.
                If False, add labor cost to all facilities in this factory.
                Defaults to True.
            add_zero_to_all_facilities (bool, optional):
                If True, add 0 labor cost to all facilities in this factory.
                If False, calculate labor cost normally.
                Defaults to False.

        Returns:
            float: Total labor cost of this factory in this time.
        """
        cost_this_time = 0.0

        if add_zero_to_all_facilities:
            for facility in self.facility_list:
                facility.cost_list.append(0.0)

        else:

            if only_working:
                for facility in self.facility_list:
                    if facility.state == BaseFacilityState.WORKING:
                        facility.cost_list.append(facility.cost_per_time)
                        cost_this_time += facility.cost_per_time
                    else:
                        facility.cost_list.append(0.0)

            else:
                for facility in self.facility_list:
                    facility.cost_list.append(facility.cost_per_time)
                    cost_this_time += facility.cost_per_time

        self.cost_list.append(cost_this_time)
        return cost_this_time

    def record_assigned_task_id(self):
        """
        Record assigned task id in this time.
        """
        for resource in self.facility_list:
            resource.record_assigned_task_id()

    def record_placed_component_id(self):
        """
        Record component id in this time.
        """
        record = []
        if len(self.placed_component_list) > 0:
            # print([c for c in self.placed_component_list])
            record = [c.ID for c in self.placed_component_list]

        self.placed_component_id_record.append(record)

    def __str__(self):
        """
        Returns:
            str: name of BaseFactory
        Examples:
            >>> factory = BaseFactory("factory")
            >>> print(factory)
            'factory'
        """
        return "{}".format(self.name)

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
    ):
        """
        Create data for gantt plotly from start_time_list and finish_time_list
        of BaseFacility in facility_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
        Returns:
            List[dict]: Gantt plotly information of this BaseFactory
        """
        df = []
        for facility in self.facility_list:
            for start_time, finish_time in zip(
                facility.start_time_list, facility.finish_time_list
            ):
                df.append(
                    dict(
                        Task=self.name + ": " + facility.name,
                        Start=(init_datetime + start_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        Finish=(
                            init_datetime
                            + (finish_time + finish_margin) * unit_timedelta
                        ).strftime("%Y-%m-%d %H:%M:%S"),
                        Type="Facility",
                    )
                )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        save_fig_path=None,
    ):
        """
        Method for creating Gantt chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(Facility="rgb(46, 137, 205)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        colors = colors if colors is not None else dict(Facility="rgb(46, 137, 205)")
        index_col = index_col if index_col is not None else "Type"
        df = self.create_data_for_gantt_plotly(init_datetime, unit_timedelta)
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            # plotly.io.write_image(fig, save_fig_path)
            print("--- Sorry, save fig is not implemented now.---")

        return fig

    def create_data_for_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
    ):
        """
        Create data for cost history plotly from cost_list in facility_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation

        Returns:
            data (List[go.Bar(name, x, y)]: Information of cost history chart.
        """
        data = []
        x = [
            (init_datetime + time * unit_timedelta).strftime("%Y-%m-%d %H:%M:%S")
            for time in range(len(self.cost_list))
        ]
        for facility in self.facility_list:
            data.append(go.Bar(name=facility.name, x=x, y=facility.cost_list))
        return data

    def create_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Cost Chart",
        save_fig_path=None,
    ):
        """
        Method for creating cost chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of cost chart.
                Defaults to "Cost Chart".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        data = self.create_data_for_cost_history_plotly(init_datetime, unit_timedelta)
        fig = go.Figure(data)
        fig.update_layout(barmode="stack", title=title)
        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
import uuid
import datetime
from .base_task import BaseTaskState


class BaseComponent(object, metaclass=abc.ABCMeta):
    """BaseComponent
    BaseComponent class for expressing target product.
    This class can be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this component.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.

        name (str):
            Basic p Defaults to None -> str(uuid.uuid4()).
        parent_component_list(List[BaseComponent], optional):
            Basic parameter.
            List of parent BaseComponents.
            Defaults to None -> [].
        child_component_list (List[BaseComponent], optional):
            Basic parameter.
            List of child BaseComponents.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted tasks.
            Defaults to None -> [].
        space_size (float, optional):
            Basic parameter.
            Space size related to base_factory's max_space_size.
            Default to None -> 1.0.
        placed_factory (BaseFactory, optional):
            Basic variable.
            A factory which this componetnt is placed in simulation.
            Defaults to None.
        placed_factory_id_record (List[str], optional):
            Basic variable.
            Record of placed factory ID in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        parent_component_list=None,
        child_component_list=None,
        targeted_task_list=None,
        space_size=None,
        # Basic variables
        placed_factory=None,
        placed_factory_id_record=None,
    ):

        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())

        if parent_component_list is not None:
            self.parent_component_list = parent_component_list
        else:
            self.parent_component_list = []

        if child_component_list is not None:
            self.child_component_list = child_component_list
        else:
            self.child_component_list = []

        if targeted_task_list is not None:
            self.targeted_task_list = targeted_task_list
        else:
            self.targeted_task_list = []

        if space_size is not None:
            self.space_size = space_size
        else:
            self.space_size = 1.0

        if placed_factory is not None:
            self.placed_factory = placed_factory
        else:
            self.placed_factory = None

        if placed_factory_id_record is not None:
            self.placed_factory_id_record = placed_factory_id_record
        else:
            self.placed_factory_id_record = []

    def extend_child_component_list(self, child_component_list):
        """
        Extend the list of child components

        Args:
            child_component_list (List[BaseComponent]):
                List of BaseComponents which are children of this component.
        Examples:
            >>> c = BaseComponent('c')
            >>> print([child_c.name for child_c in c.child_component_list])
            []
            >>> c.extend_child_component_list([BaseComponent('c1'),BaseComponent('c2')])
            >>> print([child_c.name for child_c in c.child_component_list])
            ['c1', 'c2']
        """
        for child_c in child_component_list:
            self.append_child_component(child_c)

    def append_child_component(self, child_component):
        """
        Append child component

        Args:
            child_component (BaseComponent):
                BaseComponent which is child of this component.
        Examples:
            >>> c = BaseComponent('c')
            >>> print([child_c.name for child_c in c.child_component_list])
            []
            >>> c1 = BaseComponent('c1')
            >>> c.append_child_component(c1)
            >>> print([child_c.name for child_c in c.child_component_list])
            ['c1']
            >>> print([parent_c.name for parent_c in c1.parent_component_list])
            ['c']
        """
        self.child_component_list.append(child_component)
        child_component.parent_component_list.append(self)

    def set_placed_factory(self, placed_factory, set_to_all_children=True):
        """
        Set the placed_factory

        Args:
            placed_factory (BaseFactory):
                Factory placed in this component
            set_to_all_children (bool):
                If True, set placed_factory to all children components
                Default to True
        """
        self.placed_factory = placed_factory

        if set_to_all_children:
            for child_c in self.child_component_list:
                child_c.set_placed_factory(
                    placed_factory, set_to_all_children=set_to_all_children
                )

    def is_ready(self):
        """
        Check READY component or not.
        READY component is defined by satisfying the following conditions:

        - All tasks are not NONE.
        - There is no WORKING task in this component.
        - The states of append_targeted_task includes READY.
        """
        all_none_flag = all(
            [task.state == BaseTaskState.NONE for task in self.targeted_task_list]
        )

        any_working_flag = any(
            [task.state == BaseTaskState.WORKING for task in self.targeted_task_list]
        )

        any_ready_flag = any(
            [task.state == BaseTaskState.READY for task in self.targeted_task_list]
        )

        all_finished_flag = all(
            [task.state == BaseTaskState.FINISHED for task in self.targeted_task_list]
        )

        if all_finished_flag:
            return False

        if not all_none_flag and (not any_working_flag) and any_ready_flag:
            return True

        return False

    def extend_targeted_task_list(self, targeted_task_list):
        """
        Extend the list of targeted tasks

        Args:
            targeted_task_list (List[BaseTask]):
                List of targeted tasks
        Examples:
            >>> c = BaseComponent('c')
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            []
            >>> c.extend_targeted_task_list([BaseTask('t1'),BaseTask('t2')])
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            ['t1', 't2']
        """
        for targeted_task in targeted_task_list:
            self.append_targeted_task(targeted_task)

    def append_targeted_task(self, targeted_task):
        """
        Append targeted task

        Args:
            targeted_task (BaseTask):
                Targeted task of this component
        Examples:
            >>> c = BaseComponent('c')
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            []
            >>> t1 = BaseTask('t1')
            >>> c.append_targeted_task(t1)
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            ['t1']
            >>> print(t1.target_component.name)
            'c'
        """
        self.targeted_task_list.append(targeted_task)
        targeted_task.target_component = self

    def initialize(self):
        """
        Initialize the following changeable basic variables of BaseComponent.

        - placed_factory
        - placed_factory_id_record
        """
        self.placed_factory = None
        self.placed_factory_id_record = []

    def record_placed_factory_id(self):
        """
        Record factory id in this time to placed_factory_id_record.
        """
        record = None
        if self.placed_factory is not None:
            record = self.placed_factory.ID
        self.placed_factory_id_record.append(record)

    def __str__(self):
        """
        Returns:
            str: name of BaseComponent
        Examples:
            >>> c = BaseComponent("c")
            >>> print(c)
            'c'
        """
        return "{}".format(self.name)

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
    ):
        """
        Create data for gantt plotly from start_time_list and finish_time_list
        of BaseTask in targeted_task_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
        Returns:
            list[dict]:
                Gantt plotly information of this BaseComponent
        """
        start_time = min(map(lambda t: min(t.start_time_list), self.targeted_task_list))
        finish_time = max(
            map(lambda t: max(t.finish_time_list), self.targeted_task_list)
        )
        df = [
            dict(
                Task=self.name,
                Start=(init_datetime + start_time * unit_timedelta).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                Finish=(
                    init_datetime + (finish_time + finish_margin) * unit_timedelta
                ).strftime("%Y-%m-%d %H:%M:%S"),
                Type="Component",
            )
        ]
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta
import datetime
import plotly.figure_factory as ff
import networkx as nx
import plotly.graph_objects as go
from .base_component import BaseComponent
from .base_task import BaseTask, BaseTaskState, BaseTaskDependency
from .base_team import BaseTeam
from .base_worker import BaseWorker
import itertools
from .base_resource import BaseResourceState
from .base_factory import BaseFactory
from .base_facility import BaseFacility, BaseFacilityState


class BaseProject(object, metaclass=ABCMeta):
    """BaseProject
    BaseProject class for expressing target project
    including product, organization and workflow.
    This class will be used as template.

    Args:
        file_path (str, optional):
            File path of this project data for reading.
            Defaults to None. (New Project)
        init_datetime (datetime.datetime, optional):
            Start datetime of project.
            Defaults to None -> datetime.datetime.now().
        unit_timedelta (datetime.timedelta, optional):
            Unit time of simulation.
            Defaults to None -> datetime.timedelta(minutes=1).
        product (BaseProduct, optional):
            BaseProduct in this project.
            Defaults to None. (New Project)
        organization (BaseOrganization, optional):
            BaseOrganization in this project.
            Defaults to None. (New Project)
        workflow (BaseWorkflow, optional):
            BaseWorkflow in this project.
            Defaults to None. (New Project)
        time (int, optional):
            Simulation time executing this method.
            Defaults to 0.
        cost_list (List[float], optional):
            Basic variable.
            History or record of this project's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        file_path=None,
        # Basic parameters
        init_datetime=None,
        unit_timedelta=None,
        # Basic variables
        product=None,
        organization=None,
        workflow=None,
        time=0,
        cost_list=None,
    ):

        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.init_datetime = (
            init_datetime if init_datetime is not None else datetime.datetime.now()
        )
        self.unit_timedelta = (
            unit_timedelta
            if unit_timedelta is not None
            else datetime.timedelta(minutes=1)
        )

        # Changeable variable on simulation
        # --
        # Basic variables
        if product is not None:
            self.product = product
        else:
            self.product = None

        if organization is not None:
            self.organization = organization
        else:
            self.organization = None

        if workflow is not None:
            self.workflow = workflow
        else:
            self.workflow = None

        if time != int(0):
            self.time = time
        else:
            self.time = int(0)

        if cost_list is not None:
            self.cost_list = cost_list
        else:
            self.cost_list = []

    def __str__(self):
        """
        Returns:
            str: time and name lists of product, organization and workflow.
        """
        return "TIME: {}\nPRODUCT\n{}\n\nORGANIZATION\n{}\n\nWORKFLOW\n{}".format(
            self.time, str(self.product), str(self.organization), str(self.workflow)
        )

    def initialize(self):
        """
        Initialize the changeable variables of this BaseProject.

        - time
        - cost_list
        - changeable variables of product
        - changeable variables of organization
        - changeable variables of workflow

        """
        self.time = 0
        self.cost_list = []
        self.product.initialize()
        self.organization.initialize()
        self.workflow.initialize()

    def simulate(
        self,
        task_performed_mode="multi-workers",
        error_tol=1e-10,
        print_debug=False,
        weekend_working=True,
        work_start_hour=None,
        work_finish_hour=None,
        max_time=10000,
        unit_time=1,
    ):
        """
        Simulation function for simulate this BaseProject.

        Args:
            task_performed_mode (str, optional):
                Mode of performed task in simulation.
                pDESy has the following options of this mode in simulation.

                - multi-workers

                Defaults to "multi-workers".
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.
            print_debug (bool, optional):
                Whether print debug is include or not
                Defaults to False.
            weekend_working (bool, optional):
                Whether worker works in weekend or not.
                Defaults to True.
            work_start_hour (int, optional):
                Starting working hour in one day .
                Defaults to None. This means workers work every time.
            work_finish_hour (int, optional):
                Finish working hour in one day .
                Defaults to None. This means workers work every time.
            max_time (int, optional):
                Max time of simulation.
                Defaults to 10000.
            unit_time (int, optional):
                Unit time of simulation.
                Defaults to 1.
        """

        if not (task_performed_mode == "multi-workers"):
            raise Exception(
                "Please check "
                "task_performed_mode"
                " which is equal to "
                "multi-workers"
                ""
            )

        # set simulation mode
        mode = 0
        if task_performed_mode == "multi-workers":
            mode = 1  # TaskPerformedBySingleTaskWorkers in pDES

        # check whether implementation or target mode simulation is finished or not
        if not (mode == 1):
            raise Exception("Sorry. This simulation mode is not yet implemented.")
        # -----------------------------------------------------------------------------

        self.initialize()

        while True:

            # 1. Check finished or not
            state_list = list(map(lambda task: task.state, self.workflow.task_list))
            if all(state == BaseTaskState.FINISHED for state in state_list):
                return

            # Error check
            if self.time >= max_time:
                raise Exception(
                    "Time Over! Please check your simulation model or increase "
                    "max_time"
                    " value"
                )

            # check now is business time or not
            working = True
            now_date_time = ""

            if (
                not weekend_working
                or work_start_hour is not None
                or work_finish_hour is not None
            ):
                now_date_time = self.init_datetime + self.time * self.unit_timedelta
                working = self.is_business_time(
                    now_date_time, weekend_working, work_start_hour, work_finish_hour
                )

            if print_debug:
                print("---")
                print(self.time, now_date_time, working)

            # 2. Allocate free resources to READY tasks
            if working:
                self.__allocate_single_task_workers(print_debug=print_debug)

            # 3. Pay cost to all resources in this time
            if working:
                cost_this_time = self.organization.add_labor_cost(only_working=True)
            else:
                cost_this_time = self.organization.add_labor_cost(
                    add_zero_to_all_workers=True, add_zero_to_all_facilities=True
                )
            self.cost_list.append(cost_this_time)

            # 4, Perform
            if working:
                if mode == 1:
                    self.__perform(print_debug=print_debug)

            # 5. Record
            self.__record(print_debug=print_debug)
            # 6. Update
            self.__update(print_debug=print_debug)

            self.time = self.time + unit_time

    def backward_simulate(
        self,
        task_performed_mode="multi-workers",
        error_tol=1e-10,
        print_debug=False,
        weekend_working=True,
        work_start_hour=None,
        work_finish_hour=None,
        max_time=10000,
        unit_time=-1,
        considering_due_time_of_tail_tasks=False,
        update_time_information_for_forward=True,
    ):
        """
        Backward Simulation function for simulate this BaseProject.

        Args:
            task_performed_mode (str, optional):
                Mode of performed task in simulation.
                pDESy has the following options of this mode in simulation.

                - single-worker
                - multi-workers

                Defaults to "multi-workers".
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.
            print_debug (bool, optional):
                Whether print debug is include or not
                Defaults to False.
            weekend_working (bool, optional):
                Whether worker works in weekend or not.
                Defaults to True.
            work_start_hour (int, optional):
                Starting working hour in one day .
                Defaults to None. This means workers work every time.
            work_finish_hour (int, optional):
                Finish working hour in one day .
                Defaults to None. This means workers work every time.
            max_time (int, optional):
                Max time of simulation.
                Defaults to 10000.
            unit_time (int, optional):
                Unit time of simulation.
                Defaults to -1.
            considering_due_time_of_tail_tasks (bool, optional):
                Consider due_time of tail tasks or not.
                Default to False.
            update_time_information_for_forward (bool, optional):
                Update time information for forward simulation after this simulation.
                Defaults to True.

        Note:
            This function is only for research and still in progress.
            Especially, this function is not suitable for simulation considering rework.
        """

        self.workflow.reverse_dependencies()

        autotask_removing_after_simulation = []
        try:
            if considering_due_time_of_tail_tasks:
                # Add dummy task for considering the difference of due_time
                tail_task_list = list(
                    filter(
                        lambda task: len(task.input_task_list) == 0,
                        self.workflow.task_list,
                    )
                )
                max_due_time = max([task.due_time for task in tail_task_list])
                for tail_task in tail_task_list:
                    if tail_task.due_time < max_due_time:
                        auto_task = BaseTask(
                            "auto",
                            auto_task=True,
                            default_work_amount=max_due_time - tail_task.due_time,
                        )
                        tail_task.append_input_task(
                            auto_task, task_dependency_mode=BaseTaskDependency.FS
                        )
                        autotask_removing_after_simulation.append(auto_task)
                        self.workflow.task_list.append(auto_task)
            if print_debug:
                print(autotask_removing_after_simulation)

            self.simulate(
                task_performed_mode=task_performed_mode,
                error_tol=error_tol,
                print_debug=print_debug,
                weekend_working=weekend_working,
                work_start_hour=work_start_hour,
                work_finish_hour=work_finish_hour,
                max_time=max_time,
                unit_time=unit_time,
            )

            if update_time_information_for_forward:
                # Change reverse result to normal result
                final_step = -self.time
                if considering_due_time_of_tail_tasks:
                    final_step = max_due_time

                # Task
                for task in self.workflow.task_list:
                    # Change start_time_list to finish_time_list
                    # finish_time_list to start_time_list
                    tmp = task.start_time_list
                    task.start_time_list = task.finish_time_list
                    task.finish_time_list = tmp
                    del tmp

                    # add (-final_step) to all time information
                    task.start_time_list = list(
                        map(lambda time: time + (final_step - 1), task.start_time_list)
                    )
                    task.finish_time_list = list(
                        map(lambda time: time + (final_step - 1), task.finish_time_list)
                    )

                # TODO we have to check and update this information
                min_start_time = 0
                if considering_due_time_of_tail_tasks:
                    min_start_time = float("inf")
                    for task in self.workflow.task_list:
                        tmp = min(task.start_time_list)
                        if tmp < min_start_time:
                            min_start_time = tmp
                for task in self.workflow.task_list:
                    task.ready_time_list = []
                    max_finish_time = min_start_time - 1
                    if len(task.output_task_list) > 0:  # still reverse in this line..
                        finish_time_list = []
                        for output_task in task.output_task_list:
                            finish_time_list.append(output_task[0].finish_time_list)
                        max_finish_time = max(finish_time_list)[0]
                    task.ready_time_list = [max_finish_time]

                # Team
                for team in self.organization.team_list:
                    # Worker
                    for worker in team.worker_list:
                        # Change start_time_list to finish_time_list
                        # finish_time_list to start_time_list
                        tmp = worker.start_time_list
                        worker.start_time_list = worker.finish_time_list
                        worker.finish_time_list = tmp
                        del tmp

                        # add (-final_step) to all time information
                        worker.start_time_list = list(
                            map(
                                lambda time: time + (final_step - 1),
                                worker.start_time_list,
                            )
                        )
                        worker.start_time_list.sort()
                        worker.finish_time_list = list(
                            map(
                                lambda time: time + (final_step - 1),
                                worker.finish_time_list,
                            )
                        )
                        worker.finish_time_list.sort()

                for factory in self.organization.factory_list:
                    # Facility
                    for resource in factory.facility_list:
                        # Change start_time_list to finish_time_list
                        # finish_time_list to start_time_list
                        tmp = resource.start_time_list
                        resource.start_time_list = resource.finish_time_list
                        resource.finish_time_list = tmp
                        del tmp

                        # add (-final_step) to all time information
                        resource.start_time_list = list(
                            map(
                                lambda time: time + (final_step - 1),
                                resource.start_time_list,
                            )
                        )
                        resource.start_time_list.sort()
                        resource.finish_time_list = list(
                            map(
                                lambda time: time + (final_step - 1),
                                resource.finish_time_list,
                            )
                        )
                        resource.finish_time_list.sort()

        except Exception as e:
            print(e)
        finally:
            for autotask in autotask_removing_after_simulation:
                for task, dependency in autotask.output_task_list:
                    task.input_task_list.remove([autotask, dependency])
                self.workflow.task_list.remove(autotask)
            self.workflow.reverse_dependencies()

    def __perform(self, print_debug=False):
        if print_debug:
            print("Allocation result in this time")
            worker_list = list(
                itertools.chain.from_iterable(
                    list(
                        map(lambda team: team.worker_list, self.organization.team_list)
                    )
                )
            )
            facility_list = list(
                itertools.chain.from_iterable(
                    list(
                        map(
                            lambda factory: factory.facility_list,
                            self.organization.factory_list,
                        )
                    )
                )
            )
            for worker in worker_list:
                print(
                    worker.name,
                    ":",
                    [assigned_task.name for assigned_task in worker.assigned_task_list],
                )
            for facility in facility_list:
                print(
                    facility.name,
                    ":",
                    [
                        assigned_task.name
                        for assigned_task in facility.assigned_task_list
                    ],
                )
            print("PERFORM")
        self.workflow.perform(self.time)

    def __record(self, print_debug=False):
        if print_debug:
            print("RECORD")
        self.workflow.record_allocated_workers_facilities_id()
        self.organization.record()
        self.product.record_placed_factory_id()

    def __update(self, print_debug=False):
        if print_debug:
            print("UPDATE")
        self.workflow.check_state(self.time, BaseTaskState.FINISHED)
        self.product.check_removing_placed_factory()
        self.workflow.check_state(self.time, BaseTaskState.READY)
        self.workflow.update_PERT_data(self.time)

    def __allocate_single_task_workers(self, print_debug=False):

        # Check free factory before setting components
        if print_debug:
            print("ALLOCATE")
            print("Factory - Component before setting components in this time")
            for factory in self.organization.factory_list:
                print(
                    factory.name
                    + ":"
                    + str([c.name for c in factory.placed_component_list])
                )

        target_factory_id_list = [f.ID for f in self.organization.factory_list]

        # A. Extract READY components
        ready_component_list = list(
            filter(lambda c: c.is_ready() is True, self.product.component_list)
        )
        if print_debug:
            print("Ready Component list before allocating")
            print([c.name for c in ready_component_list])

        # C. Decide which factory put each ready component
        # TODO component sorting or task sorting
        for ready_component in ready_component_list:
            ready_task_list = list(
                filter(
                    lambda task: task.state == BaseTaskState.READY,
                    ready_component.targeted_task_list,
                )
            )
            for ready_task in ready_task_list:
                for factory in ready_task.allocated_factory_list:
                    if factory.ID in target_factory_id_list:
                        if (
                            factory.can_put(ready_component)
                            and factory.get_total_workamount_skill(ready_task.name)
                            > 1e-10
                        ):
                            # move ready_component from None to factory
                            pre_factory = ready_component.placed_factory
                            if pre_factory is not None:
                                ready_component.set_placed_factory(None)
                                pre_factory.remove_placed_component(ready_component)
                            ready_component.set_placed_factory(factory)
                            factory.set_placed_component(ready_component)
                            break
                else:
                    continue
                break

        # Check free factory after setting components
        if print_debug:
            print("Factory - Component after setting components in this time")
            for factory in self.organization.factory_list:
                print(
                    factory.name
                    + ":"
                    + str([c.name for c in factory.placed_component_list])
                )
        # ---------------------------------------------------------------

        # 1. Get ready task and free resources
        ready_and_working_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.READY
                or task.state == BaseTaskState.WORKING,
                self.workflow.task_list,
            )
        )

        if print_debug:
            print("Ready & Working Task List")
            print(
                [
                    (rtask.name, rtask.remaining_work_amount)
                    for rtask in ready_and_working_task_list
                ]
            )

        # Candidate allocating task list (auto_task=False)
        ready_and_working_task_list = list(
            filter(lambda task: not task.auto_task, ready_and_working_task_list)
        )

        worker_list = list(
            itertools.chain.from_iterable(
                list(map(lambda team: team.worker_list, self.organization.team_list))
            )
        )

        free_worker_list = list(
            filter(lambda worker: worker.state == BaseResourceState.FREE, worker_list)
        )

        # 2. Sort ready task and free resources
        # Task: TSLACK (a task which Slack time(LS-ES) is lower has high priority)
        ready_and_working_task_list = sorted(
            ready_and_working_task_list, key=lambda task: task.lst - task.est
        )

        # 3. Allocate ready tasks to free resources
        for task in ready_and_working_task_list:

            # Worker: SSP
            # a resource which amount of skillpoint is lower has high priority
            free_worker_list = sorted(
                free_worker_list,
                key=lambda worker: sum(worker.workamount_skill_mean_map.values()),
            )

            allocating_workers = list(
                filter(
                    lambda worker: worker.has_workamount_skill(task.name)
                    and self.__is_allocated_worker(worker, task),
                    free_worker_list,
                )
            )

            if task.need_facility:

                # Search candidate facilities from the list of placed_factory
                placed_factory = task.target_component.placed_factory

                if placed_factory is not None:

                    free_facility_list = list(
                        filter(
                            lambda facility: facility.state == BaseFacilityState.FREE,
                            placed_factory.facility_list,
                        )
                    )

                    # Facility sorting
                    # SSP: Resource which amount of point is lower has high priority
                    free_facility_list = sorted(
                        free_facility_list,
                        key=lambda facility: sum(
                            facility.workamount_skill_mean_map.values()
                        ),
                    )

                    # candidate facilities
                    allocating_facilities = list(
                        filter(
                            lambda facility: facility.has_workamount_skill(task.name)
                            and self.__is_allocated_facility(facility, task),
                            free_facility_list,
                        )
                    )

                    for facility in allocating_facilities:
                        for worker in allocating_workers:
                            if task.can_add_resources(worker=worker, facility=facility):
                                task.allocated_worker_list.append(worker)
                                task.allocated_facility_list.append(facility)
                                allocating_workers.remove(worker)
                                free_worker_list = [
                                    w for w in free_worker_list if w.ID != worker.ID
                                ]
                                allocating_facilities.remove(facility)
                                break

            else:
                for worker in allocating_workers:
                    if task.can_add_resources(worker=worker):
                        task.allocated_worker_list.append(worker)
                        free_worker_list.remove(worker)

        # 4. Update state of task newly allocated resources (READY -> WORKING)
        self.workflow.check_state(self.time, BaseTaskState.WORKING)

    def __is_allocated_worker(self, worker, task):
        team = list(
            filter(lambda team: team.ID == worker.team_id, self.organization.team_list)
        )[0]
        return task in team.targeted_task_list

    def __is_allocated_facility(self, facility, task):
        factory = list(
            filter(
                lambda factory: factory.ID == facility.factory_id,
                self.organization.factory_list,
            )
        )[0]
        return task in factory.targeted_task_list

    def is_business_time(
        self,
        target_datetime: datetime.datetime,
        weekend_working=True,
        work_start_hour=None,
        work_finish_hour=None,
    ):
        """
        Check whether target_datetime is business time or not in this project.

        Args:
            target_datetime (datetime.datetime):
                Target datetime of checking business time or not.
            weekend_working (bool, optional):
                Whether worker works in weekend or not.
                Defaults to True.
            work_start_hour (int, optional):
                Starting working hour in one day .
                Defaults to None. This means workers work every time.
            work_finish_hour (int, optional):
                Finish working hour in one day .
                Defaults to None. This means workers work every time.

        Returns:
            bool: whether target_datetime is business time or not.
        """
        if not weekend_working:
            if target_datetime.weekday() >= 5:
                return False
            else:
                if work_start_hour is not None and work_finish_hour is not None:
                    if (
                        target_datetime.hour >= work_start_hour
                        and target_datetime.hour <= work_finish_hour
                    ):
                        return True
                    else:
                        return False
                else:
                    return True

        else:
            if work_start_hour is not None and work_finish_hour is not None:
                if (
                    target_datetime.hour >= work_start_hour
                    and target_datetime.hour <= work_finish_hour
                ):
                    return True
                else:
                    return False
            else:
                return True

    def create_gantt_plotly(
        self,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=False,
        show_colorbar=True,
        finish_margin=1.0,
        save_fig_path=None,
    ):
        """
        Method for creating Gantt chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(Component="rgb(246, 37, 105)",Task="rgb(146, 237, 5)"
                ,Worker="rgb(46, 137, 205)"
                ,Facility="rgb(46, 137, 205)",).
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else dict(
                Component="rgb(246, 37, 105)",
                Task="rgb(146, 237, 5)",
                Worker="rgb(46, 137, 205)",
                Facility="rgb(46, 137, 205)",
            )
        )
        index_col = index_col if index_col is not None else "Type"
        df = []
        df.extend(
            self.product.create_data_for_gantt_plotly(
                self.init_datetime, self.unit_timedelta, finish_margin=finish_margin
            )
        )
        df.extend(
            self.workflow.create_data_for_gantt_plotly(
                self.init_datetime, self.unit_timedelta, finish_margin=finish_margin
            ),
        )
        df.extend(
            self.organization.create_data_for_gantt_plotly(
                self.init_datetime, self.unit_timedelta, finish_margin=finish_margin
            )
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
        #     plotly.io.write_image(fig, save_fig_path)

        return fig

    def get_networkx_graph(self, view_workers=False, view_facilities=False):
        """
        Get the information of networkx graph.

        Args:
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
        Returns:
            G: networkx.Digraph()
        """
        Gp = self.product.get_networkx_graph()
        Gw = self.workflow.get_networkx_graph()
        Go = self.organization.get_networkx_graph(
            view_workers=view_workers, view_facilities=view_facilities
        )
        G = nx.compose_all([Gp, Gw, Go])

        # add edge between product and workflow
        for c in self.product.component_list:
            for task in c.targeted_task_list:
                G.add_edge(c, task)

        # add edge between workflow and team in organization
        for team in self.organization.team_list:
            for task in team.targeted_task_list:
                G.add_edge(team, task)

        if view_workers:
            for team in self.organization.team_list:
                for w in team.worker_list:
                    # G.add_node(w)
                    G.add_edge(team, w)

        # add edge between workflow and factory in organization
        for factory in self.organization.factory_list:
            for task in factory.targeted_task_list:
                G.add_edge(factory, task)

        if view_facilities:
            for factory in self.organization.factory_list:
                for w in factory.facility_list:
                    # G.add_node(w)
                    G.add_edge(factory, w)

        return G

    def draw_networkx(
        self,
        G=None,
        pos=None,
        arrows=True,
        with_labels=True,
        component_node_color="#FF6600",
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        view_facilities=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        **kwds,
    ):
        """
        Draw networkx

        Args:
            G (networkx.SDigraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            arrows (bool, optional):
                Digraph or Graph(no arrows).
                Defaults to True.
            with_labels (bool, optional):
                Label is describing or not.
                Defaults to True.
            component_node_color (str, optional):
                Node color setting information.
                Defaults to "#FF6600".
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            **kwds:
                another networkx settings.
        Returns:
            figure: Figure for a network
        """

        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        # Product
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=self.product.component_list,
            node_color=component_node_color,
        )
        # Workflow
        normal_task_list = [
            task for task in self.workflow.task_list if not task.auto_task
        ]
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=normal_task_list,
            node_color=task_node_color,
        )
        auto_task_list = [task for task in self.workflow.task_list if task.auto_task]
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=auto_task_list,
            node_color=auto_task_node_color,
        )
        # Organization - Team
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=self.organization.team_list,
            node_color=team_node_color,
            # **kwds,
        )
        if view_workers:

            worker_list = []
            for team in self.organization.team_list:
                worker_list.extend(team.worker_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                with_labels=with_labels,
                nodelist=worker_list,
                node_color=worker_node_color,
                # **kwds,
            )

        # Organization - Factory
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=self.organization.factory_list,
            node_color=factory_node_color,
            # **kwds,
        )
        if view_facilities:

            facility_list = []
            for factory in self.organization.factory_list:
                facility_list.extend(factory.facility_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                with_labels=with_labels,
                nodelist=facility_list,
                node_color=facility_node_color,
                # **kwds,
            )

        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos)

    def get_node_and_edge_trace_for_plotly_network(
        self,
        G=None,
        pos=None,
        node_size=20,
        component_node_color="#FF6600",
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            component_node_color (str, optional):
                Node color setting information.
                Defaults to "#FF6600".
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.

        Returns:
            component_node_trace: component nodes information of plotly network.
            task_node_trace: task nodes information of plotly network.
            auto_task_node_trace: auto task nodes information of plotly network.
            team_node_trace: team nodes information of plotly network.
            worker_node_trace: resource nodes information of plotly network.
            factory_node_trace: Factory Node information of plotly network.
            facility_node_trace: Facility Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        component_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=component_node_color,
                size=node_size,
            ),
        )

        task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=task_node_color,
                size=node_size,
            ),
        )

        auto_task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=auto_task_node_color,
                size=node_size,
            ),
        )

        team_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=team_node_color,
                size=node_size,
            ),
        )

        worker_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=worker_node_color,
                size=node_size,
            ),
        )

        factory_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=factory_node_color,
                size=node_size,
            ),
        )

        facility_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=facility_node_color,
                size=node_size,
            ),
        )

        edge_trace = go.Scatter(
            x=[], y=[], line=dict(width=0, color="#888"), hoverinfo="none", mode="lines"
        )

        for node in G.nodes:
            x, y = pos[node]
            if isinstance(node, BaseComponent):
                component_node_trace["x"] = component_node_trace["x"] + (x,)
                component_node_trace["y"] = component_node_trace["y"] + (y,)
                component_node_trace["text"] = component_node_trace["text"] + (node,)
            elif isinstance(node, BaseTask):
                if not node.auto_task:
                    task_node_trace["x"] = task_node_trace["x"] + (x,)
                    task_node_trace["y"] = task_node_trace["y"] + (y,)
                    task_node_trace["text"] = task_node_trace["text"] + (node,)
                elif node.auto_task:
                    auto_task_node_trace["x"] = auto_task_node_trace["x"] + (x,)
                    auto_task_node_trace["y"] = auto_task_node_trace["y"] + (y,)
                    auto_task_node_trace["text"] = auto_task_node_trace["text"] + (
                        node,
                    )
            elif isinstance(node, BaseFactory):
                factory_node_trace["x"] = factory_node_trace["x"] + (x,)
                factory_node_trace["y"] = factory_node_trace["y"] + (y,)
                factory_node_trace["text"] = factory_node_trace["text"] + (node,)
            elif isinstance(node, BaseFacility):
                facility_node_trace["x"] = facility_node_trace["x"] + (x,)
                facility_node_trace["y"] = facility_node_trace["y"] + (y,)
                facility_node_trace["text"] = facility_node_trace["text"] + (node,)
            elif isinstance(node, BaseTeam):
                team_node_trace["x"] = team_node_trace["x"] + (x,)
                team_node_trace["y"] = team_node_trace["y"] + (y,)
                team_node_trace["text"] = team_node_trace["text"] + (node,)
            elif isinstance(node, BaseWorker):
                worker_node_trace["x"] = worker_node_trace["x"] + (x,)
                worker_node_trace["y"] = worker_node_trace["y"] + (y,)
                worker_node_trace["text"] = worker_node_trace["text"] + (node,)

        for edge in G.edges:
            x = edge[0]
            y = edge[1]
            xposx, xposy = pos[x]
            yposx, yposy = pos[y]
            edge_trace["x"] += (xposx, yposx)
            edge_trace["y"] += (xposy, yposy)

        return (
            component_node_trace,
            task_node_trace,
            auto_task_node_trace,
            team_node_trace,
            worker_node_trace,
            factory_node_trace,
            facility_node_trace,
            edge_trace,
        )

    def draw_plotly_network(
        self,
        G=None,
        pos=None,
        title="Project",
        node_size=20,
        component_node_color="#FF6600",
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
        save_fig_path=None,
    ):
        """
        Draw plotly network

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            title (str, optional):
                Figure title of this network.
                Defaults to "Project".
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            component_node_color (str, optional):
                Node color setting information.
                Defaults to "#FF6600".
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a network

        TODO:
            Saving figure file is not implemented...
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)
        (
            component_node_trace,
            task_node_trace,
            auto_task_node_trace,
            team_node_trace,
            worker_node_trace,
            factory_node_trace,
            facility_node_trace,
            edge_trace,
        ) = self.get_node_and_edge_trace_for_plotly_network(G, pos, node_size=node_size)
        fig = go.Figure(
            data=[
                edge_trace,
                component_node_trace,
                task_node_trace,
                auto_task_node_trace,
                team_node_trace,
                worker_node_trace,
                factory_node_trace,
                facility_node_trace,
            ],
            layout=go.Layout(
                title=title,
                showlegend=False,
                #         hovermode='closest',
                #         margin=dict(b=20,l=5,r=5,t=40),
                annotations=[
                    dict(
                        ax=edge_trace["x"][index * 2],
                        ay=edge_trace["y"][index * 2],
                        axref="x",
                        ayref="y",
                        x=edge_trace["x"][index * 2 + 1],
                        y=edge_trace["y"][index * 2 + 1],
                        xref="x",
                        yref="y",
                        showarrow=True,
                        arrowhead=5,
                    )
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            ),
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
        #     plotly.io.write_image(fig, save_fig_path)

        return fig

    # ---
    # READ FUNCTION
    # ---

    # def read_pDESy_web_json(self, file_path: str, encoding=None):
    #     """
    #     Read json file from pDESy.org.
    #     This method is not stable now.

    #     Args:
    #         file_path (str):
    #             file path by getting pDESy.org
    #         encoding ([type], optional):
    #             Defaults to None -> utf-8.
    #     TODO:
    #         pDESy.org for describing project in web browser should be developed...
    #     """
    #     encoding = encoding if encoding is not None else "utf-8"
    #     pdes_json = open(file_path, "r", encoding=encoding)
    #     data = json.load(pdes_json)

    #     # Get Product information including Components without dependency
    #     cd_list = list(filter(lambda node: node["type"] == "Component", data))
    #     component_list = [
    #         BaseComponent(
    #             cd["name"],
    #             ID=cd["id"],
    #             error_tolerance=float(cd["userData"]["errorTolerance"]),
    #         )
    #         for cd in cd_list
    #     ]

    #     # Get Workflow information including Tasks without dependency
    #     td_list = list(filter(lambda node: node["type"] == "Task", data))
    #     task_list = [
    #         BaseTask(
    #             td["name"],
    #             ID=td["id"],
    #             default_work_amount=float(td["userData"]["workAmount"]),
    #             default_progress=float(td["userData"]["progress"]),
    #             additional_work_amount=float(td["userData"]["additionalWorkAmount"]),
    #         )
    #         for td in td_list
    #     ]

    #     # Get Organization information including Teams without dependency
    #     team_list = []
    #     ted_list = list(filter(lambda node: node["type"] == "Team", data))
    #     for team_data in ted_list:
    #         worker_list = []
    #         worker_list_data = team_data["userData"]["WorkerList"]
    #         if type(worker_list_data["Worker"]) is dict:
    #             worker_list_data["Worker"] = [worker_list_data["Worker"]]
    #         for worker_data in worker_list_data["Worker"]:
    #             work_amount_skill_mean_info = {}
    #             work_amount_skill_sd_info = {}
    #             quality_skill_mean_info = {}
    #             # quality_skill_sd_info = {}
    #             if "WorkAmountSkill" in worker_data:
    #                 if type(worker_data["WorkAmountSkill"]) is list:
    #                     for skill_data in worker_data["WorkAmountSkill"]:
    #                         work_amount_skill_mean_info[skill_data["-name"]] = float(
    #                             skill_data["-value"]
    #                         )
    #                         work_amount_skill_sd_info[skill_data["-name"]] = float(
    #                             skill_data["-value_sd"]
    #                         )
    #                 elif type(worker_data["WorkAmountSkill"]) is dict:
    #                     work_amount_skill_mean_info[
    #                         worker_data["WorkAmountSkill"]["-name"]
    #                     ] = float(worker_data["WorkAmountSkill"]["-value"])
    #                     work_amount_skill_sd_info[
    #                         worker_data["WorkAmountSkill"]["-name"]
    #                     ] = float(worker_data["WorkAmountSkill"]["-value_sd"])
    #             # if "QualitySkill" in worker_data:
    #             #     if type(worker_data["QualitySkill"]) is list:
    #             #         for skill_data in worker_data["QualitySkill"]:
    #             #             quality_skill_mean_info[skill_data["-name"]] = float(
    #             #                 skill_data["-value"]
    #             #             )
    #             #             # quality_skill_sd_info[skill_data['-name']] =
    #             #             # float(skill_data['-value_sd'])
    #             #     elif type(worker_data["QualitySkill"]) is dict:
    #             #         quality_skill_mean_info[
    #             #             worker_data["QualitySkill"]["-name"]
    #             #         ] = float(worker_data["QualitySkill"]["-value"])
    #             worker_list.append(
    #                 BaseResource(
    #                     worker_data["Name"],
    #                     team_id=team_data["id"],
    #                     cost_per_time=float(worker_data["Cost"]),
    #                     workamount_skill_mean_map=work_amount_skill_mean_info,
    #                     workamount_skill_sd_map=work_amount_skill_sd_info,
    #                     quality_skill_mean_map=quality_skill_mean_info,
    #                 )
    #             )
    #         team_list.append(
    #             BaseTeam(team_data["name"],ID=team_data["id"],worker_list=worker_list)
    #         )

    #     # Get Links information including
    #     # ComponentLink, TaskLink, TeamLink(yet), TargetComponentLink, AllocationLink
    #     l_list = list(filter(lambda node: node["type"] == "draw2d.Connection", data))
    #     for link in l_list:
    #         org_id = link["source"]["node"]
    #         org_type = list(filter(lambda node: node["id"]==org_id, data))[0]["type"]
    #         dst_id = link["target"]["node"]
    #         dst_type = list(filter(lambda node: node["id"]==dst_id, data))[0]["type"]
    #         if org_type == "Component" and dst_type == "Component":
    #             org_c = list(filter(lambda c: c.ID == org_id, component_list))[0]
    #             dst_c = list(filter(lambda c: c.ID == dst_id, component_list))[0]
    #             org_c.parent_component_list.append(dst_c)
    #             dst_c.child_component_list.append(org_c)
    #         elif org_type == "Task" and dst_type == "Task":
    #             org_task = list(filter(lambda c: c.ID == org_id, task_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == dst_id, task_list))[0]
    #             org_task.output_task_list.append(dst_task)
    #             dst_task.input_task_list.append(org_task)
    #         elif org_type == "Team" and dst_type == "Team":
    #             org_team = list(filter(lambda c: c.ID == org_id, team_list))[0]
    #             dst_team = list(filter(lambda c: c.ID == dst_id, team_list))[0]
    #             # org_task.output_task_id_list.append(dst_task.ID)
    #             dst_team.parent_team_id = org_team.ID
    #         elif org_type == "Component" and dst_type == "Task":
    #             org_c = list(filter(lambda c: c.ID == org_id, component_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == dst_id, task_list))[0]
    #             org_c.targeted_task_list.append(dst_task)
    #             dst_task.target_component_list.append(org_c)
    #         elif org_type == "Team" and dst_type == "Task":
    #             org_team = list(filter(lambda c: c.ID == org_id, team_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == dst_id, task_list))[0]
    #             org_team.targeted_task_list.append(dst_task)
    #             dst_task.allocated_team_list.append(org_team)

    #     # Aggregate
    #     self.product = BaseProduct(component_list)
    #     self.workflow = BaseWorkflow(task_list)
    #     self.organization = BaseOrganization(team_list)

    # def read_pDES_json(self, file_path: str, encoding=None):
    #     """
    #     Read json file converted from pDES file.
    #     This method is not stable now.

    #     Args:
    #         file_path (str):
    #             file path by getting pDES and converted to json
    #         encoding ([type], optional):
    #             Defaults to None -> utf-8.
    #     """
    #     encoding = encoding if encoding is not None else "utf-8"
    #     pdes_json = open(file_path, "r", encoding=encoding)
    #     data = json.load(pdes_json)

    #     # Get Product information including Components without dependency
    #     cd_list = data["ProjectDiagram"]["NodeElementList"]["ComponentNode"]
    #     component_list = [
    #         BaseComponent(
    #             cd["Name"], ID=cd["-id"],
    #         )
    #         for cd in cd_list
    #     ]

    #     # Get Workflow information including Tasks without dependency
    #     td_list = data["ProjectDiagram"]["NodeElementList"]["TaskNode"]
    #     task_list = [
    #         BaseTask(
    #             td["Name"],
    #             ID=td["-id"],
    #             default_work_amount=float(td["WorkAmount"]),
    #             default_progress=float(td["Progress"]),
    #             # additional_work_amount=float(td["AdditionalWorkAmount"]),
    #         )
    #         for td in td_list
    #     ]

    #     # Get Organization information including Teams without dependency
    #     team_list = []
    #     ted_list = data["ProjectDiagram"]["NodeElementList"]["TeamNode"]
    #     for team_data in ted_list:
    #         worker_list = []
    #         worker_list_data = team_data["WorkerList"]
    #         if type(worker_list_data["Worker"]) is dict:
    #             worker_list_data["Worker"] = [worker_list_data["Worker"]]
    #         for worker_data in worker_list_data["Worker"]:
    #             work_amount_skill_mean_info = {}
    #             work_amount_skill_sd_info = {}
    #             quality_skill_mean_info = {}
    #             # quality_skill_sd_info = {}
    #             if "WorkAmountSkill" in worker_data:
    #                 if type(worker_data["WorkAmountSkill"]) is list:
    #                     for skill_data in worker_data["WorkAmountSkill"]:
    #                         work_amount_skill_mean_info[skill_data["-name"]] = float(
    #                             skill_data["-value"]
    #                         )
    #                         work_amount_skill_sd_info[skill_data["-name"]] = float(
    #                             skill_data["-value_sd"]
    #                         )
    #                 elif type(worker_data["WorkAmountSkill"]) is dict:
    #                     work_amount_skill_mean_info[
    #                         worker_data["WorkAmountSkill"]["-name"]
    #                     ] = float(worker_data["WorkAmountSkill"]["-value"])
    #                     work_amount_skill_sd_info[
    #                         worker_data["WorkAmountSkill"]["-name"]
    #                     ] = float(worker_data["WorkAmountSkill"]["-value_sd"])
    #             # if "QualitySkill" in worker_data:
    #             #     if type(worker_data["QualitySkill"]) is list:
    #             #         for skill_data in worker_data["QualitySkill"]:
    #             #             quality_skill_mean_info[skill_data["-name"]] = float(
    #             #                 skill_data["-value"]
    #             #             )
    #             #             # quality_skill_sd_info[skill_data['-name']]
    #             #             #  = float(skill_data['-value_sd'])
    #             #     elif type(worker_data["QualitySkill"]) is dict:
    #             #         quality_skill_mean_info[
    #             #             worker_data["QualitySkill"]["-name"]
    #             #         ] = float(worker_data["QualitySkill"]["-value"])
    #             worker_list.append(
    #                 BaseResource(
    #                     worker_data["Name"],
    #                     team_id=team_data["-id"],
    #                     cost_per_time=float(worker_data["Cost"]),
    #                     workamount_skill_mean_map=work_amount_skill_mean_info,
    #                     workamount_skill_sd_map=work_amount_skill_sd_info,
    #                     quality_skill_mean_map=quality_skill_mean_info,
    #                 )
    #             )
    #         team_list.append(
    #             BaseTeam(
    #                 team_data["Name"], ID=team_data["-id"], worker_list=worker_list
    #             )
    #         )
    #     self.organization = BaseOrganization(team_list)

    #     # Get Links information including
    #     # ComponentLink, TaskLink, TeamLink(yet), TargetComponentLink, AllocationLink
    #     l_list = data["ProjectDiagram"]["LinkList"]["Link"]
    #     for link in l_list:
    #         if link["-type"] == "ComponentLink":
    #             org_c = list(filter(lambda c: c.ID==link["-org"], component_list))[0]
    #             dst_c = list(filter(lambda c: c.ID==link["-dst"], component_list))[0]
    #             org_c.parent_component_list.append(dst_c)
    #             dst_c.child_component_list.append(org_c)
    #         elif link["-type"] == "TaskLink":
    #             org_task = list(filter(lambda c: c.ID == link["-org"], task_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == link["-dst"], task_list))[0]
    #             org_task.output_task_list.append(dst_task)
    #             dst_task.input_task_list.append(org_task)
    #         elif link["-type"] == "TeamLink":
    #             org_team = list(filter(lambda c: c.ID == link["-org"], team_list))[0]
    #             dst_team = list(filter(lambda c: c.ID == link["-dst"], team_list))[0]
    #             # org_task.output_task_id_list.append(dst_task.ID)
    #             dst_team.get_work_amount_skill_progress_team_id = org_team.ID
    #         elif link["-type"] == "TargetComponentLink":
    #             org_c = list(filter(lambda c: c.ID==link["-org"], component_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == link["-dst"], task_list))[0]
    #             org_c.targeted_task_list.append(dst_task)
    #             dst_task.target_component_list.append(org_c)
    #         elif link["-type"] == "AllocationLink":
    #             org_team = list(filter(lambda c: c.ID == link["-org"], team_list))[0]
    #             dst_task = list(filter(lambda c: c.ID == link["-dst"], task_list))[0]
    #             org_team.targeted_task_list.append(dst_task)
    #             dst_task.allocated_team_list.append(org_team)

    #     # Aggregate
    #     self.product = BaseProduct(component_list)
    #     self.workflow = BaseWorkflow(task_list)
    #     self.organization = BaseOrganization(team_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base_component import BaseComponent
from .base_product import BaseProduct
from typing import List


class Product(BaseProduct):
    """BaseProduct
    Product class for expressing target product in a project.
    This class is implemented from BaseProduct.

    Args:
        component_list (List[BaseComponent]):
            List of BaseComponents
    """

    def __init__(self, component_list: List[BaseComponent]):
        super().__init__(component_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
from typing import List
from .base_task import BaseTask, BaseTaskState, BaseTaskDependency
from .base_resource import BaseResourceState
import plotly.figure_factory as ff
import networkx as nx
import plotly.graph_objects as go
import datetime
import matplotlib.pyplot as plt


class BaseWorkflow(object, metaclass=abc.ABCMeta):
    """BaseWorkflow
    BaseWorkflow class for expressing workflow in a project.
    BaseWorkflow is consist of multiple BaseTasks.
    This class will be used as template.

    Args:
        task_list (List[BaseTask]):
            Basic parameter.
            List of BaseTask in this BaseWorkflow.
        critical_path_length (float, optional):
            Basic variable.
            Critical path length of PERT/CPM.
            Defaults to 0.0.
    """

    def __init__(
        self,
        # Basic parameters
        task_list: List[BaseTask],
        # Basic variables
        critical_path_length=0.0,
    ):
        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.task_list = task_list

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        self.critical_path_length = (
            critical_path_length if critical_path_length != 0.0 else 0.0
        )

    def __str__(self):
        """
        Returns:
            str: name list of BaseTask
        Examples:
            >>> w = BaseWorkflow([BaseTask('t1')])
            >>> print([t.name for t in w.task_list])
            ['t1']
        """
        return "{}".format(list(map(lambda task: str(task), self.task_list)))

    def initialize(self):
        """
        Initialize the changeable variables of BaseWorkflow including PERT calculation.

        - changeable variables of BaseTask in task_list
        - critical_path_length
        """
        for task in self.task_list:
            task.initialize()
        self.critical_path_length = 0.0
        self.update_PERT_data(0)
        self.check_state(-1, BaseTaskState.READY)

    def record_allocated_workers_facilities_id(self):
        for task in self.task_list:
            task.record_allocated_workers_facilities_id()

    def update_PERT_data(self, time: int):
        """
        Update PERT data (est,eft,lst,lft) of each BaseTask in task_list

        Args:
            time (int):
                Simulation time.
        """
        self.__set_est_eft_data(time)
        self.__set_lst_lft_criticalpath_data(time)

    def check_state(self, time: int, state: BaseTaskState):
        """
        Check state of all BaseTasks in task_list

        Args:
            time (int):
                Simulation time
            state (BaseTaskState):
                Check target state.
                Search and update all tasks which can change only target state.
        """
        if state == BaseTaskState.READY:
            self.__check_ready(time)
        elif state == BaseTaskState.WORKING:
            self.__check_working(time)
        elif state == BaseTaskState.FINISHED:
            self.__check_finished(time)

    def __check_ready(self, time: int):
        none_task_list = list(
            filter(lambda task: task.state == BaseTaskState.NONE, self.task_list)
        )
        for none_task in none_task_list:
            input_task_list = none_task.input_task_list

            # check READY condition by each dependency
            # FS: if input task is finished
            # SS: if input task is started
            # ...or this is head task
            ready = True
            for input_task, dependency in input_task_list:
                if dependency == BaseTaskDependency.FS:
                    if input_task.state == BaseTaskState.FINISHED:
                        ready = True
                    else:
                        ready = False
                        break
                elif dependency == BaseTaskDependency.SS:
                    if input_task.state == BaseTaskState.WORKING:
                        ready = True
                    else:
                        ready = False
                        break
                elif dependency == BaseTaskDependency.SF:
                    pass
                elif dependency == BaseTaskDependency.FF:
                    pass
            if ready:
                none_task.state = BaseTaskState.READY
                none_task.ready_time_list.append(time)

    def __check_working(self, time: int):
        ready_and_assigned_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.READY
                and len(task.allocated_worker_list) > 0,
                self.task_list,
            )
        )

        ready_auto_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.READY and task.auto_task,
                self.task_list,
            )
        )

        working_and_assigned_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.WORKING
                and len(task.allocated_worker_list) > 0,
                self.task_list,
            )
        )

        target_task_list = []
        target_task_list.extend(ready_and_assigned_task_list)
        target_task_list.extend(ready_auto_task_list)
        target_task_list.extend(working_and_assigned_task_list)

        for task in target_task_list:
            if task.state == BaseTaskState.READY:
                task.state = BaseTaskState.WORKING
                task.start_time_list.append(time)
                for worker in task.allocated_worker_list:
                    worker.state = BaseResourceState.WORKING
                    worker.start_time_list.append(time)
                    worker.assigned_task_list.append(task)
                if task.need_facility:
                    for facility in task.allocated_facility_list:
                        facility.state = BaseResourceState.WORKING
                        facility.start_time_list.append(time)
                        facility.assigned_task_list.append(task)

            elif task.state == BaseTaskState.WORKING:
                for worker in task.allocated_worker_list:
                    if worker.state == BaseResourceState.FREE:
                        worker.state = BaseResourceState.WORKING
                        worker.start_time_list.append(time)
                        worker.assigned_task_list.append(task)
                    if task.need_facility:
                        for facility in task.allocated_facility_list:
                            if facility.state == BaseResourceState.FREE:
                                facility.state = BaseResourceState.WORKING
                                facility.start_time_list.append(time)
                                facility.assigned_task_list.append(task)

    def __check_finished(self, time: int, error_tol=1e-10):
        working_and_zero_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.WORKING
                and task.remaining_work_amount < 0.0 + error_tol,
                self.task_list,
            )
        )
        for task in working_and_zero_task_list:
            # check FINISH condition by each dependency
            # SF: if input task is working
            # FF: if input task is finished
            finished = True
            for input_task, dependency in task.input_task_list:
                if dependency == BaseTaskDependency.FS:
                    pass
                elif dependency == BaseTaskDependency.SS:
                    pass
                elif dependency == BaseTaskDependency.SF:
                    if input_task.state == BaseTaskState.WORKING:
                        finished = True
                    else:
                        finished = False
                        break
                elif dependency == BaseTaskDependency.FF:
                    if input_task.state == BaseTaskState.FINISHED:
                        finished = True
                    else:
                        finished = False
                        break
            if finished:
                task.state = BaseTaskState.FINISHED
                task.finish_time_list.append(time)
                task.remaining_work_amount = 0.0

                for worker in task.allocated_worker_list:
                    if all(
                        list(
                            map(
                                lambda task: task.state == BaseTaskState.FINISHED,
                                worker.assigned_task_list,
                            )
                        )
                    ):
                        worker.state = BaseResourceState.FREE
                        worker.finish_time_list.append(time)
                        worker.assigned_task_list.remove(task)
                task.allocated_worker_list = []

                if task.need_facility:
                    for facility in task.allocated_facility_list:
                        if all(
                            list(
                                map(
                                    lambda task: task.state == BaseTaskState.FINISHED,
                                    facility.assigned_task_list,
                                )
                            )
                        ):
                            facility.state = BaseResourceState.FREE
                            facility.finish_time_list.append(time)
                            facility.assigned_task_list.remove(task)

                    task.allocated_facility_list = []

    def __set_est_eft_data(self, time: int):

        input_task_list = []

        # 1. Set the earliest finish time of head tasks.
        for task in self.task_list:
            task.est = time
            if len(task.input_task_list) == 0:
                task.eft = time + task.remaining_work_amount
                input_task_list.append(task)

        # 2. Calculate PERT information of all tasks
        while len(input_task_list) > 0:
            next_task_list = []
            for input_task in input_task_list:
                for next_task, dependency in input_task.output_task_list:
                    pre_est = next_task.est
                    est = 0
                    eft = 0
                    if dependency == BaseTaskDependency.FS:
                        est = input_task.est + input_task.remaining_work_amount
                        eft = est + next_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.SS:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.FF:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                        if input_task.eft > eft:
                            eft = input_task.eft
                    elif dependency == BaseTaskDependency.SF:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                        if input_task.est > eft:
                            eft = input_task.est
                    else:
                        est = input_task.est + input_task.remaining_work_amount
                        eft = est + next_task.remaining_work_amount
                    if est >= pre_est:
                        next_task.est = est
                        next_task.eft = eft
                    next_task_list.append(next_task)

            input_task_list = next_task_list

    def __set_lst_lft_criticalpath_data(self, time: int):

        # 1. Extract the list of tail tasks.
        output_task_list = list(
            filter(lambda task: len(task.output_task_list) == 0, self.task_list)
        )

        # 2. Update the information of critical path of this workflow.
        self.critical_path_length = max(output_task_list, key=lambda task: task.eft).eft
        for task in output_task_list:
            task.lft = self.critical_path_length
            task.lst = task.lft - task.remaining_work_amount

        # 3. Calculate PERT information of all tasks
        while len(output_task_list) > 0:

            prev_task_list = []
            for output_task in output_task_list:
                for prev_task, dependency in output_task.input_task_list:
                    pre_lft = prev_task.lft
                    lst = 0
                    lft = 0
                    if dependency == BaseTaskDependency.FS:
                        lft = output_task.lst
                        lst = lft - prev_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.SS:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.FF:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                        if output_task.lft < lft:
                            lft = output_task.lft
                    elif dependency == BaseTaskDependency.SF:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                        if output_task.lft < lst:
                            lst = output_task.lft
                    else:
                        lft = output_task.lst
                        lst = lft - prev_task.remaining_work_amount
                    if pre_lft < 0 or pre_lft >= lft:
                        prev_task.lst = lst
                        prev_task.lft = lft
                    prev_task_list.append(prev_task)

            output_task_list = prev_task_list

    def reverse_dependencies(self):
        """
        Reverse all task dependencies in task_list.

        Note:
            This method is developed only for backward simulation.
        """
        # 1.
        # Register the input_task_list to dummy_output_task_list
        # Register the output_task_list to dummy_input_task_list
        for task in self.task_list:
            task.dummy_output_task_list = task.input_task_list
            task.dummy_input_task_list = task.output_task_list

        # 2.
        # Register the dummy_output_task_list to output_task_list
        # Register the dummy_input_task_list to input_task_list
        # Delete the dummy_output_task_list, dummy_input_task_list
        for task in self.task_list:
            task.output_task_list = task.dummy_output_task_list
            task.input_task_list = task.dummy_input_task_list
            del task.dummy_output_task_list, task.dummy_input_task_list

    def perform(self, time: int, seed=None):
        """
        Perform BaseTask in task_list in simulation.

        Args:
            time (int):
                Simulation time.
            seed (int, optional):
                Random seed for describing deviation of progress.
                If workamount
                Defaults to None.
        Note:
            This method includes advanced code of custom simulation.
            We have to separete basic code and advanced code in the future.
        """
        for task in self.task_list:
            task.perform(time, seed=seed)

    def create_simple_gantt(
        self,
        finish_margin=1.0,
        view_auto_task=False,
        view_ready=False,
        task_color="#00EE00",
        auto_task_color="#005500",
        ready_color="#C0C0C0",
        save_fig_path=None,
    ):
        """

        Method for creating Gantt chart by matplotlib.
        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_auto_task (bool, optional):
                View auto_task or not.
                Defaults to False.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            task_color (str, optional):
                Task color setting information.
                Defaults to "#00EE00".
            auto_task_color (str, optional):
                Auto Task color setting information.
                Defaults to "#005500".
            ready_color (str, optional):
                Ready Task color setting information.
                Defaults to "#C0C0C0".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
            gnt: gnt in plt.subplots()
        """
        fig, gnt = plt.subplots()
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_task_list = list(filter(lambda task: not task.auto_task, self.task_list))
        if view_auto_task:
            target_task_list = self.task_list

        yticks = [10 * (n + 1) for n in range(len(target_task_list))]
        yticklabels = [task.name for task in target_task_list]

        gnt.set_yticks(yticks)
        gnt.set_yticklabels(yticklabels)

        for ttime in range(len(target_task_list)):
            task = target_task_list[ttime]
            wlist = []  # time list from start to finish+finish_margin
            rlist = []  # time list from ready to start
            for wtime in range(len(task.start_time_list)):
                wlist.append(
                    (
                        task.start_time_list[wtime],
                        task.finish_time_list[wtime]
                        - task.start_time_list[wtime]
                        + finish_margin,
                    )
                )
                rlist.append(
                    (
                        task.ready_time_list[wtime] + finish_margin,
                        task.start_time_list[wtime] - task.ready_time_list[wtime],
                    )
                )
            if task.auto_task:
                if view_ready:
                    gnt.broken_barh(
                        rlist, (yticks[ttime] - 5, 9), facecolors=(ready_color)
                    )
                gnt.broken_barh(
                    wlist, (yticks[ttime] - 5, 9), facecolors=(auto_task_color)
                )
            else:
                if view_ready:
                    gnt.broken_barh(
                        rlist, (yticks[ttime] - 5, 9), facecolors=(ready_color)
                    )
                gnt.broken_barh(wlist, (yticks[ttime] - 5, 9), facecolors=(task_color))

        if save_fig_path is not None:
            plt.savefig(save_fig_path)

        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
        view_ready=False,
    ):
        """
        Create data for gantt plotly
        from task_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.

        Returns:
            list[dict]: Gantt plotly information of this BaseWorkflow
        """
        df = []
        for task in self.task_list:
            df.extend(
                task.create_data_for_gantt_plotly(
                    init_datetime,
                    unit_timedelta,
                    finish_margin=finish_margin,
                    view_ready=view_ready,
                )
            )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        finish_margin=1.0,
        view_ready=False,
        save_fig_path=None,
    ):
        """
        Method for creating Gantt chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None ->
                dict(WORKING="rgb(146, 237, 5)", READY="rgb(107, 127, 135)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "State".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else dict(WORKING="rgb(146, 237, 5)", READY="rgb(107, 127, 135)")
        )
        index_col = index_col if index_col is not None else "State"
        df = self.create_data_for_gantt_plotly(
            init_datetime,
            unit_timedelta,
            finish_margin=finish_margin,
            view_ready=view_ready,
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
        #     plotly.io.write_image(fig, save_fig_path)

        return fig

    def get_networkx_graph(self):
        """
        Get the information of networkx graph.

        Returns:
            G: networkx.Digraph()
        """
        G = nx.DiGraph()

        # 1. add all nodes
        for task in self.task_list:
            G.add_node(task)

        # 2. add all edges
        for task in self.task_list:
            for input_task, dependency in task.input_task_list:
                G.add_edge(input_task, task)

        return G

    def draw_networkx(
        self,
        G=None,
        pos=None,
        arrows=True,
        with_labels=True,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        **kwds,
    ):
        """
        Draw networkx

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            arrows (bool, optional):
                Digraph or Graph(no arrows).
                Defaults to True.
            with_labels (bool, optional):
                Label is describing or not.
                Defaults to True.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            **kwds:
                another networkx settings.
        Returns:
            figure: Figure for a network
        """
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)
        # nx.draw_networkx(G, pos=pos, arrows=arrows, with_labels=with_labels, **kwds)

        # normal task
        normal_task_list = [task for task in self.task_list if not task.auto_task]
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=normal_task_list,
            node_color=task_node_color,
            # **kwds,
        )

        # auto task
        auto_task_list = [task for task in self.task_list if task.auto_task]
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=auto_task_list,
            node_color=auto_task_node_color,
            # **kwds,
        )

        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos)

    def get_node_and_edge_trace_for_plotly_network(
        self,
        G=None,
        pos=None,
        node_size=20,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".

        Returns:
            task_node_trace: Normal Task Node information of plotly network.
            auto_task_node_trace: Auto Task Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)

        task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=task_node_color,
                size=node_size,
            ),
        )

        auto_task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=auto_task_node_color,
                size=node_size,
            ),
        )

        for node in G.nodes:
            x, y = pos[node]
            if not node.auto_task:
                task_node_trace["x"] = task_node_trace["x"] + (x,)
                task_node_trace["y"] = task_node_trace["y"] + (y,)
                task_node_trace["text"] = task_node_trace["text"] + (node,)
            elif node.auto_task:
                auto_task_node_trace["x"] = auto_task_node_trace["x"] + (x,)
                auto_task_node_trace["y"] = auto_task_node_trace["y"] + (y,)
                auto_task_node_trace["text"] = auto_task_node_trace["text"] + (node,)

        edge_trace = go.Scatter(
            x=[], y=[], line=dict(width=1, color="#888"), hoverinfo="none", mode="lines"
        )

        for edge in G.edges:
            x = edge[0]
            y = edge[1]
            xposx, xposy = pos[x]
            yposx, yposy = pos[y]
            edge_trace["x"] += (xposx, yposx)
            edge_trace["y"] += (xposy, yposy)

        return task_node_trace, auto_task_node_trace, edge_trace

    def draw_plotly_network(
        self,
        G=None,
        pos=None,
        title="Workflow",
        node_size=20,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        save_fig_path=None,
    ):
        """
        Draw plotly network

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            title (str, optional):
                Figure title of this network.
                Defaults to "Workflow".
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a network

        TODO:
            Saving figure file is not implemented...
        """
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)
        (
            task_node_trace,
            auto_task_node_trace,
            edge_trace,
        ) = self.get_node_and_edge_trace_for_plotly_network(
            G,
            pos,
            node_size=node_size,
            task_node_color=task_node_color,
            auto_task_node_color=auto_task_node_color,
        )
        fig = go.Figure(
            data=[edge_trace, task_node_trace, auto_task_node_trace],
            layout=go.Layout(
                title=title,
                showlegend=False,
                #         hovermode='closest',
                #         margin=dict(b=20,l=5,r=5,t=40),
                annotations=[
                    dict(
                        ax=edge_trace["x"][index * 2],
                        ay=edge_trace["y"][index * 2],
                        axref="x",
                        ayref="y",
                        x=edge_trace["x"][index * 2 + 1],
                        y=edge_trace["y"][index * 2 + 1],
                        xref="x",
                        yref="y",
                        showarrow=True,
                        arrowhead=5,
                    )
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            ),
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
        #     plotly.io.write_image(fig, save_fig_path)

        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
from typing import List
from .base_team import BaseTeam
from .base_factory import BaseFactory
from .base_worker import BaseWorker
from .base_facility import BaseFacility
import plotly.figure_factory as ff
import networkx as nx
import plotly.graph_objects as go
import datetime
import matplotlib.pyplot as plt


class BaseOrganization(object, metaclass=abc.ABCMeta):
    """BaseOrganization
    BaseOrganization class for expressing organization in target project.
    BaseOrganization is consist of multiple BaseTeam and BaseFactory.
    This class will be used as template.

    Args:
        team_list (List[BaseTeam]):
            Basic parameter.
            List of BaseTeam in this organization.
        factory_list (List[BaseFactory], optional):
            Basic parameter.
            List of BaseFactory in this organization.
            Default to None -> []
        cost_list (List[float], optional):
            Basic variable.
            History or record of this organization's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        team_list: List[BaseTeam],
        factory_list=None,
        # Basic variables
        cost_list=None,
    ):
        # ----
        # Constraint parameters on simulation
        # --
        # Basic parameter
        self.team_list = team_list
        self.factory_list = factory_list if factory_list is not None else []
        # --
        # Advanced parameter for customized simulation

        # ----
        # Changeable variables on simulation
        # --
        # Basic variables
        self.cost_list = []
        # --
        # Advanced variables for customized simulation

    def __str__(self):
        """
        Returns:
            str: name list of BaseTeam
        Examples:
            >>> o = BaseOrganization([BaseTeam('t1')])
            >>> print([t.name for t in o.team_list])
            ['t1']
        """
        return "{}".format(list(map(lambda team: str(team), self.team_list)))

    def initialize(self):
        """
        Initialize the changeable variables of BaseOrganization

        - cost_list
        - changeable variables of BaseTeam in team_list

        """
        self.cost_list = []
        for team in self.team_list:
            team.initialize()
        for factory in self.factory_list:
            factory.initialize()

    def add_labor_cost(
        self,
        only_working=True,
        add_zero_to_all_workers=False,
        add_zero_to_all_facilities=False,
    ):
        """
        Add labor cost to teams and workers in this organization.

        Args:
            only_working (bool, optional):
                If True, add labor cost to only WORKING workers in this organization.
                If False, add labor cost to all workers in this organization.
                Defaults to True.
            add_zero_to_all_workers (bool, optional):
                If True, add 0 labor cost to all workers in this team.
                If False, calculate labor cost normally.
                Defaults to False.
            add_zero_to_all_facilities (bool, optional):
                If True, add 0 labor cost to all facilities in this team.
                If False, calculate labor cost normally.
                Defaults to False.
        Returns:
            float: Total labor cost of this team in this time.
        """
        cost_this_time = 0.0
        for team in self.team_list:
            cost_this_time += team.add_labor_cost(
                only_working=only_working,
                add_zero_to_all_workers=add_zero_to_all_workers,
            )
        for factory in self.factory_list:
            cost_this_time += factory.add_labor_cost(
                only_working=only_working,
                add_zero_to_all_facilities=add_zero_to_all_facilities,
            )
        self.cost_list.append(cost_this_time)
        return cost_this_time

    def record(self):
        """
        Record assigned task id and component in this time.
        """
        for team in self.team_list:
            team.record_assigned_task_id()
        for factory in self.factory_list:
            factory.record_assigned_task_id()
            factory.record_placed_component_id()

    def create_simple_gantt(
        self,
        finish_margin=1.0,
        view_workers=True,
        view_facilities=True,
        team_color="#0099FF",
        worker_color="#D9E5FF",
        factory_color="#0099FF",
        facility_color="#D9E5FF",
        save_fig_path=None,
    ):
        """

        Method for creating Gantt chart by matplotlib.
        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to True.
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to True.
            team_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            factory_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
            gnt: gnt in plt.subplots()
        Note:
            view_worker=False mode is not implemented now...

        TODO:
            view_worker=False mode should be implemented.
            view_facility=False mode should be implemented.
        """
        fig, gnt = plt.subplots()
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_worker_list = []
        target_facility_list = []
        yticklabels = []
        for team in self.team_list:
            for worker in team.worker_list:
                target_worker_list.append(worker)
                yticklabels.append(team.name + ":" + worker.name)
        for factory in self.factory_list:
            for facility in factory.facility_list:
                target_facility_list.append(facility)
                yticklabels.append(factory.name + ":" + facility.name)
        yticks = [
            10 * (n + 1)
            for n in range(len(target_worker_list) + len(target_facility_list))
        ]

        gnt.set_yticks(yticks)
        gnt.set_yticklabels(yticklabels)

        for ttime in range(len(target_worker_list)):
            worker = target_worker_list[ttime]
            wlist = []
            for wtime in range(len(worker.start_time_list)):
                wlist.append(
                    (
                        worker.start_time_list[wtime],
                        worker.finish_time_list[wtime]
                        - worker.start_time_list[wtime]
                        + finish_margin,
                    )
                )
            gnt.broken_barh(wlist, (yticks[ttime] - 5, 9), facecolors=(worker_color))

        for ttime in range(len(target_facility_list)):
            facility = target_facility_list[ttime]
            wlist = []
            for wtime in range(len(facility.start_time_list)):
                wlist.append(
                    (
                        facility.start_time_list[wtime],
                        facility.finish_time_list[wtime]
                        - facility.start_time_list[wtime]
                        + finish_margin,
                    )
                )
            gnt.broken_barh(
                wlist,
                (yticks[ttime + len(target_worker_list)] - 5, 9),
                facecolors=(worker_color),
            )

        if save_fig_path is not None:
            plt.savefig(save_fig_path)

        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
    ):
        """
        Create data for gantt plotly from team_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
        Returns:
            List[dict]: Gantt plotly information of this BaseOrganization.
        """
        df = []
        for team in self.team_list:
            df.extend(
                team.create_data_for_gantt_plotly(
                    init_datetime, unit_timedelta, finish_margin=finish_margin
                )
            )
        for factory in self.factory_list:
            df.extend(
                factory.create_data_for_gantt_plotly(
                    init_datetime, unit_timedelta, finish_margin=finish_margin
                )
            )
        return df

    def create_gantt_plotly(
        self,
        init_datetime,
        unit_timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        finish_margin=1.0,
        save_fig_path=None,
    ):
        """
        Method for creating Gantt chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(Worker="rgb(46, 137, 205)",Facility="rgb(46, 137, 205)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else dict(Worker="rgb(46, 137, 205)", Facility="rgb(46, 137, 205)")
        )
        index_col = index_col if index_col is not None else "Type"
        df = self.create_data_for_gantt_plotly(
            init_datetime, unit_timedelta, finish_margin=finish_margin
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            #     plotly.io.write_image(fig, save_fig_path)
            print("--- Sorry, save fig is not implemented now.---")

        return fig

    def create_data_for_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
    ):
        """
        Create data for cost history plotly from cost_list of team_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation

        Returns:
            data (List[go.Bar(name, x, y)]: Information of cost history chart.
        """
        data = []
        x = [
            (init_datetime + time * unit_timedelta).strftime("%Y-%m-%d %H:%M:%S")
            for time in range(len(self.cost_list))
        ]
        for team in self.team_list:
            data.append(go.Bar(name=team.name, x=x, y=team.cost_list))
        for factory in self.factory_list:
            data.append(go.Bar(name=factory.name, x=x, y=factory.cost_list))
        return data

    def create_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Cost Chart",
        save_fig_path=None,
    ):
        """
        Method for creating cost chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of cost chart.
                Defaults to "Cost Chart".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        data = self.create_data_for_cost_history_plotly(init_datetime, unit_timedelta)
        fig = go.Figure(data)
        fig.update_layout(barmode="stack", title=title)
        return fig

    def get_networkx_graph(self, view_workers=False, view_facilities=False):
        """
        Get the information of networkx graph.

        Args:
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
        Returns:
            G: networkx.Digraph()
        """
        G = nx.DiGraph()

        # BaseTeam
        # 1. add all nodes
        for team in self.team_list:
            G.add_node(team)

        # 2. add all edges
        for team in self.team_list:
            if team.parent_team is not None:
                G.add_edge(team.parent_team, team)

        if view_workers:
            for team in self.team_list:
                for w in team.worker_list:
                    G.add_node(w)
                    G.add_edge(team, w)

        # BaseFactory
        # 1. add all nodes
        for factory in self.factory_list:
            G.add_node(factory)

        # 2. add all edges
        for factory in self.factory_list:
            if factory.parent_factory is not None:
                G.add_edge(factory.parent_factory, factory)

        if view_facilities:
            for factory in self.factory_list:
                for w in factory.facility_list:
                    G.add_node(w)
                    G.add_edge(factory, w)

        return G

    def draw_networkx(
        self,
        G=None,
        pos=None,
        arrows=True,
        with_labels=True,
        view_workers=False,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_facilities=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        **kwds,
    ):
        """
        Draw networkx

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            arrows (bool, optional):
                Digraph or Graph(no arrows).
                Defaults to True.
            with_labels (bool, optional):
                Label is describing or not.
                Defaults to True.
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            **kwds:
                another networkx settings.
        Returns:
            figure: Figure for a network
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        # nx.draw_networkx(G, pos=pos, arrows=arrows, with_labels=with_labels, **kwds)

        # team
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=self.team_list,
            node_color=team_node_color,
            # **kwds,
        )
        # resources
        if view_workers:

            worker_list = []
            for team in self.team_list:
                worker_list.extend(team.worker_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                with_labels=with_labels,
                nodelist=worker_list,
                node_color=worker_node_color,
                # **kwds,
            )

        # factory
        nx.draw_networkx_nodes(
            G,
            pos,
            with_labels=with_labels,
            nodelist=self.factory_list,
            node_color=factory_node_color,
            # **kwds,
        )
        # facility
        if view_facilities:

            facility_list = []
            for factory in self.factory_list:
                facility_list.extend(factory.facility_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                with_labels=with_labels,
                nodelist=facility_list,
                node_color=facility_node_color,
                # **kwds,
            )

        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos)

    def get_node_and_edge_trace_for_plotly_network(
        self,
        G=None,
        pos=None,
        node_size=20,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.

        Returns:
            team_node_trace: Team Node information of plotly network.
            worker_node_trace: Worker Node information of plotly network.
            factory_node_trace: Factory Node information of plotly network.
            facility_node_trace: Facility Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """

        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        team_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=team_node_color,
                size=node_size,
            ),
        )
        worker_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=worker_node_color,
                size=node_size,
            ),
        )
        factory_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=factory_node_color,
                size=node_size,
            ),
        )
        facility_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=facility_node_color,
                size=node_size,
            ),
        )

        for node in G.nodes:
            x, y = pos[node]
            if isinstance(node, BaseTeam):
                team_node_trace["x"] = team_node_trace["x"] + (x,)
                team_node_trace["y"] = team_node_trace["y"] + (y,)
                team_node_trace["text"] = team_node_trace["text"] + (node,)
            elif isinstance(node, BaseFactory):
                factory_node_trace["x"] = factory_node_trace["x"] + (x,)
                factory_node_trace["y"] = factory_node_trace["y"] + (y,)
                factory_node_trace["text"] = factory_node_trace["text"] + (node,)
            elif isinstance(node, BaseFacility):
                facility_node_trace["x"] = facility_node_trace["x"] + (x,)
                facility_node_trace["y"] = facility_node_trace["y"] + (y,)
                facility_node_trace["text"] = facility_node_trace["text"] + (node,)
            elif isinstance(node, BaseWorker):
                worker_node_trace["x"] = worker_node_trace["x"] + (x,)
                worker_node_trace["y"] = worker_node_trace["y"] + (y,)
                worker_node_trace["text"] = worker_node_trace["text"] + (node,)

        edge_trace = go.Scatter(
            x=[], y=[], line=dict(width=1, color="#888"), hoverinfo="none", mode="lines"
        )

        for edge in G.edges:
            x = edge[0]
            y = edge[1]
            xposx, xposy = pos[x]
            yposx, yposy = pos[y]
            edge_trace["x"] += (xposx, yposx)
            edge_trace["y"] += (xposy, yposy)

        return (
            team_node_trace,
            worker_node_trace,
            factory_node_trace,
            facility_node_trace,
            edge_trace,
        )

    def draw_plotly_network(
        self,
        G=None,
        pos=None,
        title="Organization",
        node_size=20,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        factory_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
        save_fig_path=None,
    ):
        """
        Draw plotly network

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            title (str, optional):
                Figure title of this network.
                Defaults to "Organization".
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            factory_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a network

        TODO:
            Saving figure file is not implemented...
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)
        (
            team_node_trace,
            worker_node_trace,
            factory_node_trace,
            facility_node_trace,
            edge_trace,
        ) = self.get_node_and_edge_trace_for_plotly_network(
            G,
            pos,
            node_size=node_size,
            team_node_color=team_node_color,
            worker_node_color=worker_node_color,
            view_workers=view_workers,
            factory_node_color=factory_node_color,
            facility_node_color=facility_node_color,
            view_facilities=view_facilities,
        )
        fig = go.Figure(
            data=[
                edge_trace,
                team_node_trace,
                worker_node_trace,
                factory_node_trace,
                facility_node_trace,
            ],
            layout=go.Layout(
                title=title,
                showlegend=False,
                #         hovermode='closest',
                #         margin=dict(b=20,l=5,r=5,t=40),
                annotations=[
                    dict(
                        ax=edge_trace["x"][index * 2],
                        ay=edge_trace["y"][index * 2],
                        axref="x",
                        ayref="y",
                        x=edge_trace["x"][index * 2 + 1],
                        y=edge_trace["y"][index * 2 + 1],
                        xref="x",
                        yref="y",
                        showarrow=True,
                        arrowhead=5,
                    )
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            ),
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
        #     plotly.io.write_image(fig, save_fig_path)

        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base_resource import BaseResource, BaseResourceState


class BaseWorker(BaseResource):
    """BaseWorker
    BaseResource class for expressing a worker.
    This class will be used as template.
    This class is implemented from BaseResource.

    Args:
        name (str):
            Basic parameter. Name of this resource.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None.
        factory_id (str, optional):
            Basic parameter.
            Factory ID will be defined automatically on adding factory.
            Defaults to None.
        cost_per_time (float, optional):
            Basic parameter.
            Cost of this resource per unit time.
            Defaults to 0.0.
        solo_working (bool, optional):
            Basic parameter.
            Flag whether this resource can work with other resources or not.
            Defaults to False.
        workamount_skill_mean_map (Dict[str, float], optional):
            Basic parameter.
            Skill for expressing progress in unit time.
            Defaults to {}.
        workamount_skill_sd_map (Dict[str, float], optional):
            Basic parameter.
            Standard deviation of skill for expressing progress in unit time.
            Defaults to {}.
        facility_skill_map (Dict[str, float], optional):
            Basic parameter.
            Skill for operating facility in unit time.
            Defaults to {}.
        state (BaseResourceState, optional):
            Basic variable.
            State of this resource in simulation.
            Defaults to BaseResourceState.FREE.
        cost_list (List[float], optional):
            Basic variable.
            History or record of his or her cost in simulation.
            Defaults to None -> [].
        start_time_list (List[int], optional):
            Basic variable.
            History or record of his or her start time in simulation.
            Defaults to None -> [].
        finish_time_list (List[int], optional):
            Basic variable.
            History or record of his or her finish time in simulation.
            Defaults to None -> [].
        assigned_task_list (List[BaseTask], optional):
            Basic variable.
            State of his or her assigned tasks in simulation.
            Defaults to None -> [].
        assigned_task_id_record (List[List[str]], optional):
            Basic variable.
            Record of his or her assigned tasks' id in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        team_id=None,
        cost_per_time=0.0,
        solo_working=False,
        workamount_skill_mean_map={},
        workamount_skill_sd_map={},
        facility_skill_map={},
        # Basic variables
        state=BaseResourceState.FREE,
        cost_list=None,
        start_time_list=None,
        finish_time_list=None,
        assigned_task_list=None,
        assigned_task_id_record=None,
    ):
        super().__init__(
            name,
            ID=ID,
            team_id=team_id,
            cost_per_time=cost_per_time,
            solo_working=solo_working,
            workamount_skill_mean_map=workamount_skill_mean_map,
            workamount_skill_sd_map=workamount_skill_sd_map,
            state=state,
            cost_list=cost_list,
            start_time_list=start_time_list,
            finish_time_list=finish_time_list,
            assigned_task_list=assigned_task_list,
            assigned_task_id_record=assigned_task_id_record,
        )

        self.facility_skill_map = (
            facility_skill_map if facility_skill_map is not None else {}
        )

    def has_facility_skill(self, facility_name, error_tol=1e-10):
        """
        Check whether he or she has facility skill or not
        by checking facility_skill_map.

        Args:
            facility_name (str):
                Facility name
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.

        Returns:
            bool: whether he or she has workamount skill of task_name or not
        """
        if facility_name in self.facility_skill_map:
            if self.facility_skill_map[facility_name] > 0.0 + error_tol:
                return True
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
import uuid
from .base_resource import BaseResourceState
import plotly.graph_objects as go
import plotly.figure_factory as ff
import datetime


class BaseTeam(object, metaclass=abc.ABCMeta):
    """
    BaseTeam class for expressing team in a project.
    This class will be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this team.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None.
        worker_list (List[BaseWorker], optional):
            Basic parameter.
            List of BaseWorkers who belong to this team.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted BaseTasks.
            Defaults to None -> [].
        parent_team (BaseTeam, optional):
            Basic parameter.
            Parent team of this team.
            Defaults to None.
        cost_list (List[float], optional):
            Basic variable.
            History or record of this team's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        worker_list=None,
        targeted_task_list=None,
        parent_team=None,
        # Basic variables
        cost_list=None,
    ):

        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())

        self.worker_list = worker_list if worker_list is not None else []
        for worker in self.worker_list:
            if worker.team_id is None:
                worker.team_id = self.ID

        self.targeted_task_list = (
            targeted_task_list if targeted_task_list is not None else []
        )
        self.parent_team = parent_team if parent_team is not None else None

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        if cost_list is not None:
            self.cost_list = cost_list
        else:
            self.cost_list = []

    def set_parent_team(self, parent_team):
        """
        Set parent team

        Args:
            parent_team (BaseTeam):
                Parent team
        Examples:
            >>> t = BaseTeam('t')
            >>> t1 = BaseTeam('t1')
            >>> t.set_parent_team(t1)
            >>> print(t.parent_team.name)
            't1'
        """
        self.parent_team = parent_team

    def extend_targeted_task_list(self, targeted_task_list):
        """
        Extend the list of targeted tasks

        Args:
            targeted_task_list (list[BaseTask]):
                List of targeted tasks
        Examples:
            >>> team = BaseTeam('team')
            >>> print([targeted_t.name for targeted_t in team.targeted_task_list])
            []
            >>> team.extend_targeted_task_list([BaseTask('t1'),BaseTask('t2')])
            >>> print([targeted_t.name for targeted_t in team.targeted_task_list])
            ['t1', 't2']
        """
        for targeted_task in targeted_task_list:
            self.append_targeted_task(targeted_task)

    def append_targeted_task(self, targeted_task):
        """
        Append targeted task

        Args:
            targeted_task (BaseTask):
                Targeted task
        Examples:
            >>> team = BaseTeam('team')
            >>> print([targeted_t.name for targeted_t in team.targeted_task_list])
            []
            >>> t1 = BaseTask('t1')
            >>> team.append_targeted_task(t1)
            >>> print([targeted_t.name for targeted_t in team.targeted_task_list])
            ['t1']
            >>> print([target_team.name for target_team in t1.allocated_team_list])
            ['team']
        """
        self.targeted_task_list.append(targeted_task)
        targeted_task.allocated_team_list.append(self)

    def add_worker(self, worker):
        """
        Add worker to self.worker_list

        Args:
            worker (BaseWorker):
                Worker which is added to this factory
        """
        worker.team_id = self.ID
        self.worker_list.append(worker)

    def initialize(self):
        """
        Initialize the changeable variables of BaseTeam

        - cost_list
        - changeable variable of BaseWorker in worker_list
        """
        self.cost_list = []
        for w in self.worker_list:
            w.initialize()

    def add_labor_cost(self, only_working=True, add_zero_to_all_workers=False):
        """
        Add labor cost to workers in this team.

        Args:
            only_working (bool, optional):
                If True, add labor cost to only WORKING workers in this team.
                If False, add labor cost to all workers in this team.
                Defaults to True.
            add_zero_to_all_workers (bool, optional):
                If True, add 0 labor cost to all workers in this team.
                If False, calculate labor cost normally.
                Defaults to False.

        Returns:
            float: Total labor cost of this team in this time.
        """
        cost_this_time = 0.0

        if add_zero_to_all_workers:
            for worker in self.worker_list:
                worker.cost_list.append(0.0)

        else:

            if only_working:
                for worker in self.worker_list:
                    if worker.state == BaseResourceState.WORKING:
                        worker.cost_list.append(worker.cost_per_time)
                        cost_this_time += worker.cost_per_time
                    else:
                        worker.cost_list.append(0.0)

            else:
                for worker in self.worker_list:
                    worker.cost_list.append(worker.cost_per_time)
                    cost_this_time += worker.cost_per_time

        self.cost_list.append(cost_this_time)
        return cost_this_time

    def record_assigned_task_id(self):
        """
        Record assigned task id in this time.
        """
        for worker in self.worker_list:
            worker.record_assigned_task_id()

    def __str__(self):
        """
        Returns:
            str: name of BaseTeam
        Examples:
            >>> team = BaseTeam("team")
            >>> print(team)
            'team'
        """
        return "{}".format(self.name)

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
    ):
        """
        Create data for gantt plotly from start_time_list and finish_time_list
        of BaseWorker in worker_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
        Returns:
            List[dict]: Gantt plotly information of this BaseTeam
        """
        df = []
        for worker in self.worker_list:
            for start_time, finish_time in zip(
                worker.start_time_list, worker.finish_time_list
            ):
                df.append(
                    dict(
                        Task=self.name + ": " + worker.name,
                        Start=(init_datetime + start_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        Finish=(
                            init_datetime
                            + (finish_time + finish_margin) * unit_timedelta
                        ).strftime("%Y-%m-%d %H:%M:%S"),
                        Type="Worker",
                    )
                )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        save_fig_path=None,
    ):
        """
        Method for creating Gantt chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(Worker="rgb(46, 137, 205)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        colors = colors if colors is not None else dict(Worker="rgb(46, 137, 205)")
        index_col = index_col if index_col is not None else "Type"
        df = self.create_data_for_gantt_plotly(init_datetime, unit_timedelta)
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            print("--- Sorry, save fig is not implemented now.---")
            # plotly.io.write_image(fig, save_fig_path)

        return fig

    def create_data_for_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
    ):
        """
        Create data for cost history plotly from cost_list of BaseWorker in worker_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation

        Returns:
            data (List[go.Bar(name, x, y)]: Information of cost history chart.
        """
        data = []
        x = [
            (init_datetime + time * unit_timedelta).strftime("%Y-%m-%d %H:%M:%S")
            for time in range(len(self.cost_list))
        ]
        for worker in self.worker_list:
            data.append(go.Bar(name=worker.name, x=x, y=worker.cost_list))
        return data

    def create_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Cost Chart",
        save_fig_path=None,
    ):
        """
        Method for creating cost chart by plotly.
        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of cost chart.
                Defaults to "Cost Chart".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Saving figure file is not implemented...
        """
        data = self.create_data_for_cost_history_plotly(init_datetime, unit_timedelta)
        fig = go.Figure(data)
        fig.update_layout(barmode="stack", title=title)
        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import abc
import uuid
from enum import IntEnum
import numpy as np
from .base_task import BaseTaskState


class BaseFacilityState(IntEnum):
    FREE = 0
    WORKING = 1


class BaseFacility(object, metaclass=abc.ABCMeta):
    """BaseFacility
    BaseFacility class for expressing a factory.
    This class will be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this resource.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None -> str(uuid.uuid4()).
        factory_id (str, optional):
            Basic parameter.
            ID of Factory which this facility is belonged.
            Defaults to None.
        cost_per_time (float, optional):
            Basic parameter.
            Cost of this facility per unit time.
            Defaults to 0.0.
        solo_working (bool, optional):
            Basic parameter.
            Flag whether this facility can work any task with other facilities or not.
            Defaults to False.
        workamount_skill_mean_map (Dict[str, float], optional):
            Basic parameter.
            Mean skill for expressing progress in unit time.
            Defaults to {}.
        workamount_skill_sd_map (Dict[str, float], optional):
            Basic parameter.
            Standard deviation of skill for expressing progress in unit time.
            Defaults to {}.
        state (BaseFacilityState, optional):
            Basic variable.
            State of this resource in simulation.
            Defaults to BaseFacilityState.FREE.
        cost_list (List[float], optional):
            Basic variable.
            History or record of his or her cost in simulation.
            Defaults to None -> [].
        start_time_list (List[int], optional):
            Basic variable.
            History or record of his or her start time in simulation.
            Defaults to None -> [].
        finish_time_list (List[int], optional):
            Basic variable.
            History or record of his or her finish time in simulation.
            Defaults to None -> [].
        assigned_task_list (List[BaseTask], optional):
            Basic variable.
            State of his or her assigned tasks in simulation.
            Defaults to None -> [].
        assigned_task_id_record (List[List[str]], optional):
            Basic variable.
            Record of his or her assigned tasks' id in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        factory_id=None,
        cost_per_time=0.0,
        solo_working=False,
        workamount_skill_mean_map={},
        workamount_skill_sd_map={},
        # Basic variables
        state=BaseFacilityState.FREE,
        cost_list=None,
        start_time_list=None,
        finish_time_list=None,
        assigned_task_list=None,
        assigned_task_id_record=None,
    ):

        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())
        self.factory_id = factory_id if factory_id is not None else None
        self.cost_per_time = cost_per_time if cost_per_time != 0.0 else 0.0
        self.solo_working = solo_working if solo_working is not None else False
        self.workamount_skill_mean_map = (
            workamount_skill_mean_map if workamount_skill_mean_map is not {} else {}
        )
        self.workamount_skill_sd_map = (
            workamount_skill_sd_map if workamount_skill_sd_map is not None else {}
        )

        # ----
        # Changeable variablng factory.e on simulation
        # --
        # Basic variables
        if state is not BaseFacilityState.FREE:
            self.state = state
        else:
            self.state = BaseFacilityState.FREE

        if cost_list is not None:
            self.cost_list = cost_list
        else:
            self.cost_list = []

        if start_time_list is not None:
            self.start_time_list = start_time_list
        else:
            self.start_time_list = []

        if finish_time_list is not None:
            self.finish_time_list = finish_time_list
        else:
            self.finish_time_list = []

        if assigned_task_list is not None:
            self.assigned_task_list = assigned_task_list
        else:
            self.assigned_task_list = []

        if assigned_task_id_record is not None:
            self.assigned_task_id_record = assigned_task_id_record
        else:
            self.assigned_task_id_record = []

    def __str__(self):
        """
        Returns:
            str: name of BaseResource
        Examples:
            >>> r = BaseResource("r")
            >>> print(r)
            'r'
        """
        return "{}".format(self.name)

    def initialize(self):
        """
        Initialize the following basic variables of BaseFacility

        - state
        - cost_list
        - start_time_list
        - finish_time_list
        - assigned_task_list
        - assigned_task_id_record
        """
        self.state = BaseFacilityState.FREE
        self.cost_list = []
        self.start_time_list = []
        self.finish_time_list = []
        self.assigned_task_list = []
        self.assigned_task_id_record = []

    def record_assigned_task_id(self):
        """
        Record assigned task id in this time.
        """
        self.assigned_task_id_record.append(
            [task.ID for task in self.assigned_task_list]
        )

    def has_workamount_skill(self, task_name, error_tol=1e-10):
        """
        Check whether he or she has workamount skill or not
        by checking workamount_skill_mean_map.

        Args:
            task_name (str):
                Task name
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.

        Returns:
            bool: whether he or she has workamount skill of task_name or not
        """
        if task_name in self.workamount_skill_mean_map:
            if self.workamount_skill_mean_map[task_name] > 0.0 + error_tol:
                return True
        return False

    def get_work_amount_skill_progress(self, task_name, seed=None):
        """
        Get progress of workamount by his or her contribution in this time.

        If he or she has multipul tasks in this time,
        progress `p_r(t)` is defined as follows:

        p_r(t)={ps_r(t)}/{N_r(t)}

        - `ps_r(t)`: progress if he or she has only this task in this time
        - `N_r(t)`: Number of allocated tasks to him or her in this time


        Args:
            task_name (str):
                Task name
            error_tol (float, optional):
                Countermeasures against numerical error.
                Defaults to 1e-10.

        Returns:
            float: Progress of workamount by his or her contribution in this time
        """
        if seed is not None:
            np.random.seed(seed=seed)
        if not self.has_workamount_skill(task_name):
            return 0.0
        skill_mean = self.workamount_skill_mean_map[task_name]
        if task_name not in self.workamount_skill_sd_map:
            skill_sd = 0
        else:
            skill_sd = self.workamount_skill_sd_map[task_name]
        base_progress = np.random.normal(skill_mean, skill_sd)
        sum_of_working_task_in_this_time = sum(
            map(
                lambda task: task.state == BaseTaskState.WORKING
                or task.state == BaseTaskState.WORKING_ADDITIONALLY,
                self.assigned_task_list,
            )
        )
        return base_progress / float(sum_of_working_task_in_this_time)
